#!/usr/bin/perl
# FILE: metatest
# AUTHOR: William Stafford Noble
# CREATE DATE: 7-13-97
# PROJECT: MHMM
# COPYRIGHT: 2001, Columbia University
# DESCRIPTION: Run a complete set of tests on the Meta-MEME software.

$usage = "
This script tests each program in the Meta-MEME software distribution.

USAGE: metatest [-update] [-complete] [-continue] [-notrain]

  -update : Update the files to which the script compares program outputs.
  -complete : Include in the test the scaffolding programs that are created
              with 'make tests'.
  -continue : Allow the tests to continue after a program test fails.";


# Parse the command line.
if (scalar(@ARGV) > 4) {
  print(STDERR $usage);
  exit(1);
}

# Set defaults for global variables.
$update = 0;
$complete = 0;
$continue = 0;
$train = 1;

# Read the optional parameters.
while (scalar(@ARGV) > 0) {
  $param = shift(@ARGV);
  if ($param eq "-update") {
    $update = 1;
  } elsif ($param eq "-complete") {
    $complete = 1;
  } elsif ($param eq "-continue") {
    $continue = 1;
  } elsif ($param eq "-notrain") {
    $train = 0;
  } else {
    print(STDERR "Unrecognized option ($param).\n");
    print(STDERR $usage);
    exit(1);
  }
}

# Check if the env tell us
# where to run (needed by 'make distcheck' in build process)
$testbindir = `pwd`;
chomp $testbindir;
$testbindir = $testbindir . "/../src";
if ($ENV{testdir}) {
	chdir $ENV{testdir};
	print("testdir = $ENV{testdir}\n");
}

if ($complete) {
  &test_scaffold("crp0", 1);
  &test_scaffold("lipo", 0);
}
&test_core("crp0", 1);
&test_core("lipo", 0);

# Test mhmms with --maxseqs option (r1635)
&test_program($update, $continue, "mhmms", 
  "--maxseqs 3 --quiet"
    . " mhmm/crp0.linear.mhmm"
    . " common/crp0.fasta",
    "mhmms/r1635.maxseqs.mhmms");

# Test mhmms with --global option (r1635)
&test_program($update, $continue, "mhmms", 
  "--global --quiet"
    . " mhmm/lipo.linear.mhmm"
    . " common/lipo.fasta",
    "mhmms/r1635.global.mhmms");

# Test mhmms for bug where runs of 
# spacer states are scored as matches (r1635).
# With theses input files mhmms should find no matches.
&test_program($update, $continue, "mhmms", 
  "--quiet --motif-scoring"
    . " mhmm/lipo.linear.mhmm"
    . " mhmms/r1635.fasta",
    "mhmms/r1635.mhmms");

# Test mhmms for segmentation fault when --motif-scoring option 
# is used. (r1584)
&test_program($update, $continue, "mhmms", 
  "--motif-scoring --quiet"
    . " mhmms/r1584.hmm"
    . " mhmms/r1584.fasta",
    "mhmms/r1584.mhmms");

# Test motiph
&test_program($update, $continue, "motiph", 
  "--bg 2.0 --pseudocount 0.01 --gff - "
    . " motiph/spiked.aln"
    . " motiph/yeast.tree"
    . " motiph/MCM1.meme.html",
    "motiph/motiph.gff");

# Test motiph with --motif option
&test_program($update, $continue, "motiph", 
  "--bg 2.0 --pseudocount 0.01 --gff - "
    . " --pthresh 1.0"
    . " --motif 2 --motif 3"
    . " motiph/spiked.aln"
    . " motiph/yeast.tree"
    . " common/crp0.meme.html",
    "motiph/motiph-motif23.gff");

# Test fimo
# The MEME_ETC_DIR contains the needed style sheets.
$ENV{'MEME_ETC_DIR'} = '@MEME_ETC_DIR@';
&test_program($update, $continue, "fimo", 
  "--text --motif-pseudo 0.01"
    . " motiph/MCM1.meme.html"
    . " motiph/spiked.fasta",
    "motiph/fimo.txt");

# Test fimo with --motif option
# The MEME_ETC_DIR contains the needed style sheets.
$ENV{'MEME_ETC_DIR'} = '@MEME_ETC_DIR@';
&test_program($update, $continue, "fimo", 
  "--text --motif 2 --motif 3"
    . " --output-pthresh 0.01"
    . " common/crp0.meme.html"
    . " motiph/spiked.fasta",
    "motiph/fimo-motif23.txt");

# Test shadow 
#  &test_program($update, $continue, "shadow", 
#  "--pthresh 0.01 --bg 2.0"
#    . " motiph/spiked.aln"
#    . " motiph/yeast.tree",
#    "motiph/shadow.gff");

# Test tomtom (2 distance measures)
&test_program($update, $continue, "tomtom", 
  " -query common/sample.meme -target common/sample.meme"
    . " -dist allr  -text ",
    "tomtom/tomtom.out.allr");
&test_program($update, $continue, "tomtom", 
  " -query common/sample.meme -target common/sample.meme"
    . " -dist pearson  -text ",
    "tomtom/tomtom.out.pearson");

exit(0);

###########################################################################
sub test_scaffold {
  local($fileroot, $is_dna) = @_;

  # Create input filenames.
  $train_file = "common/$fileroot.fasta";
  $meme_file = "common/$fileroot.meme.html";
  $test_file = "common/$fileroot-test.fasta";
  $linear_model = "mhmm/$fileroot.linear.mhmm";
  $complete_model = "mhmm/$fileroot.complete.mhmm";
  $test_aln = "common/test.aln";
  $test_aln_out = "clustalw2fasta/test.fasta";
  $test_aln_nogap_out = "clustalw2fasta/test.nogap.fasta";
  $test_aln_consensus_out = "clustalw2fasta/test.consensus.fasta";

  # All files are in the scaffold directory.
  $scaffold_root = "scaffold/$fileroot";

  if ($is_dna == 1) {
    $dna_switch = "--dna";
  } else {
    $dna_switch = "";
  }

  # Run a test on each program in succession.
  &test_program($update, $continue, "meme-io", " --verbosity 1 $meme_file ",
		"$scaffold_root.meme-io");
  &test_program($update, $continue, "fasta-io",
		" --verbosity 1 $dna_switch $train_file ", $train_file);
  &test_program($update, $continue, "fasta-io", 
		" --verbosity 1  --many $dna_switch $train_file", $train_file);
  &test_program($update, $continue, "fasta-io", 
		" --verbosity 1 --blocksize 100 $train_file", "$scaffold_root.fasta");

  # Test model I/O.
  &test_program($update, $continue, "mhmm-io", " --verbosity 1 $linear_model", 
		"$linear_model");
  &test_program($update, $continue, "mhmm-io", " --verbosity 1 $complete_model",
		"$complete_model");

  # Test model log conversion.
  &test_program($update, $continue, "log-hmm", " --verbosity 1 $linear_model", 
		"$linear_model");
  &test_program($update, $continue, "log-hmm", " --verbosity 1 $complete_model",
		"$complete_model");

  # Test clustalw2fasta
  &test_program($update, $continue, "clustalw2fasta", " $test_aln", 
    $test_aln_out);
  &test_program($update, $continue, "clustalw2fasta", " -nogap $test_aln", 
    $test_aln_nogap_out);
  &test_program($update, $continue, "clustalw2fasta", 
    " -consensus 100 $test_aln", $test_aln_consensus_out); 
}

###########################################################################
sub test_core {
  local($fileroot, $is_dna) = @_;

  # Create input filenames.
  $train_file = "common/$fileroot.fasta";
  $meme_file = "common/$fileroot.meme.html";
  $test_file = "common/$fileroot-test.fasta";

  # Check different topologies.
  foreach $topology ("linear", "complete", "star") {

    # Create various types of spacer models.
    foreach $spacer ("1", "3", "fim") {

      # FIXME: There is a bug with star topology plus fims.
      if (($spacer eq "fim") && ($topology eq "star")) {
				next;
      }

      $model = "$fileroot.$topology";
      $mhmm_params = " --noheader --noparams --type $topology ";

      if ($spacer eq "fim") {
				$model .= ".fim";
				$mhmm_params .= "--fim ";
      } elsif ($spacer eq "3") {
				$model .= ".spacer";
				$mhmm_params .= "--nspacer 3";
      }
			$mhmm_params .= " --verbosity 1 $meme_file";
      
      # Create the model.
      &test_program($update, $continue, "mhmm", $mhmm_params, 
		    "mhmm/$model.mhmm");

      # Draw the model.
      &test_program($update, $continue, "draw-mhmm",
		    " --verbosity 1 --consensus mhmm/$model.mhmm ", 
		    "draw-mhmm/$model.gvz");

      # Set parameters for search routines.
      $search_params = "--ethresh 99999 --quiet --width 79 ";

      # Use different scoring schemes.
      foreach $paths ("all", "single") {

				# Search with the model.	    
				$search_file = "$model.$paths";
				if ($paths ne "all") {
	  			$search_params .= "--fancy ";
				}
				&test_program($update, $continue, "mhmms", 
		      " --verbosity 1 --paths $paths $search_params mhmm/$model.mhmm $test_file",
		      "mhmms/$search_file.mhmms");
      }
    }
  }
}

############################################################################
sub test_program {
  local($update, $continue, $program, $arguments, $good_file) = @_;
  $output = "/var/tmp/$program.$$.tmp";
  
  # Tell the user what's happening.
  printf("************************************************************\n");
  printf("Testing $program . . . \n");

  # Run the program and collect the error level.
  printf("$testbindir/$program $arguments \n");
  system("$testbindir/$program $arguments > $output");
  $result = $? >> 8;

  # Eliminate any lines that might differ, but are irrelevant
  system("sed -e '/xml-stylesheet/d' $output > $output.sed");

  # Diff the results with the desired results.
  $diff = `diff $good_file $output.sed`;

  # Check for problems and die if any happened.
  if (($diff ne "") || ($result != 0) || ($? >> 8 != 0)) {
    printf("FAIL\n");
    if ($result != 0) {
      printf("$program had a non-zero exit status ($result)\n");
    } else {
      print("$good_file\n");
      print($diff);
      if ($update) {
				print(STDERR "Updating $good_file.\n");
				`cp $output.sed $good_file`;
      }
    }
    if (!$continue) {
      exit(1);
    }
  } else {

    # Return success.
    printf("SUCCESS\n");
  }

  # Delete the temporary file.
  unlink($output);
  unlink($output.sed);
}
