/********************************************************************
 * FILE: motiph.c
 * AUTHOR: William Stafford Noble, Charles E. Grant, Timothy L. Bailey
 * CREATE DATE: 12/17/2004
 * PROJECT: EVOMCAST
 * COPYRIGHT: 2004, UW
 ********************************************************************/

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include "alignment.h"
#include "alphabet.h"
#include "evomodel.h"
#include "fasta-io.h"
#include "hash_alph.h"
#include "meme-io.h"
#include "mhmm-state.h"
#include "motif.h"
#include "motiph-scoring.h"
#include "scored-sites.h"
#include "pssm.h"
#include "seq.h"
#include "simple-getopt.h"
#include "tree.h"
#include "cisml.h"
#include "object-list.h"

char* program_name = NULL;
VERBOSE_T verbosity = NORMAL_VERBOSE;

// Nucleotide alphabet order as in motif.h
extern char alphabet[];

static TREE_T* trim_tree_to_alignment(TREE_T* tree, ALIGNMENT_T* alignment);
static void print_col_frequencies(ARRAY_T* alignment_col_freqs);

/*************************************************************************
 * Entry point for motiph
 *************************************************************************/
int main(int argc, char *argv[]) {

  BOOLEAN_T scan_both_strands = TRUE;
  BOOLEAN_T use_file_list = FALSE;
  char* bg_filename = NULL;
  char* motif_name = "motif"; // Use this motif name in the output.
  char* seq_name = NULL;      // Use this sequence name in the output.
  STRING_LIST_T* selected_motifs = NULL;
  double pthresh = 1e-6;
  double gap_cost = 0.0;
  double fg_rate = 1.0;
  double bg_rate = 1.0;
  double purine_pyrimidine = 1.0; // r
  double transition_transversion = 0.5; // R
  double pseudocount = 0.1;
  BOOLEAN_T compute_qvalues = FALSE;
  char* gff_filename = NULL;
  GAP_SUPPORT_T gap_support = SKIP_GAPS;
  MODEL_TYPE_T model_type = F81_MODEL;
  COLUMN_FREQS_TYPE_T column_freqs_type = SIMULATED_COL_FREQS;
  BOOLEAN_T use_halpern_bruno = FALSE;
  BOOLEAN_T print_col_freqs = FALSE;
  BOOLEAN_T print_trimmed_tree = FALSE;
  OBJECT_LIST_T* alignment_col_freqs_list = NULL;
  char* ustar_label = NULL;	// TLB; create uniform star tree

  program_name = "motiph";

  /**********************************************
   * COMMAND LINE PROCESSING
   **********************************************/

  // Define command line options. (FIXME: Repeated code)
  // FIXME: Note that if you add or remove options you
  // must change n_options.
  int n_options = 23;
  cmdoption const motiph_options[] = {
    {"list", NO_VALUE},
    {"hb", NO_VALUE},
    {"ustar", REQUIRED_VALUE},
    {"pthresh", REQUIRED_VALUE},
    {"model", REQUIRED_VALUE},
    {"column-freqs", REQUIRED_VALUE},
    {"pur-pyr", REQUIRED_VALUE},
    {"transition-transversion", REQUIRED_VALUE},
    {"bg", REQUIRED_VALUE},
    {"fg", REQUIRED_VALUE},
    {"gap", REQUIRED_VALUE},
    {"gap-cost", REQUIRED_VALUE},
    {"motif", REQUIRED_VALUE},
    {"motif-name", REQUIRED_VALUE},
    {"bgfile", REQUIRED_VALUE},
    {"sequence-name", REQUIRED_VALUE},
    {"norc", NO_VALUE},
    {"pseudocount", REQUIRED_VALUE},
    {"print-col-freqs", NO_VALUE},
    {"print-trimmed-tree", NO_VALUE},
    {"qvalue", NO_VALUE},
    {"gff", REQUIRED_VALUE},
    {"verbosity", REQUIRED_VALUE}
  };

  int option_index = 0;

  // Define the usage message.
  char      usage[1000] = "";
  strcat(usage, "USAGE: motiph [options] <alignment> <tree> <motif>\n");
  strcat(usage, "\n");
  strcat(usage, "   Options:\n");

  // Evolutionary model parameters.
  strcat(usage, "     --list\n");
  strcat(usage, "     --hb\n");
  strcat(usage, "     --model single|average|jc|k2|f81|f84|hky|tn");
  strcat(usage, " (default=f81)\n");
  strcat(usage, "     --column-freqs simulated|empirical (default=simulated)\n");
  strcat(usage, "     --pur-pyr <float> (default=1.0)\n");
  strcat(usage, "     --transition-transversion <float> (default=0.5)\n");
  strcat(usage, "     --bg <float> (default=1.0)\n");
  strcat(usage, "     --fg <float> (default=1.0)\n");
  strcat(usage, "     --gap skip | fixed | wildcard | minimum");
  strcat(usage, " (default=skip)\n");
  strcat(usage, "     --gap-cost <float> (default=0.0)\n");

  // Motif parameters.
  strcat(usage, "     --motif <id> (default=all)\n");
  strcat(usage, "     --motif-name <string> (default from motif file)\n");

  // Miscellaneous parameters
  strcat(usage, "     --sequence-name <string> (default first sequence" \
                " in alignment)\n");
  strcat(usage, "     --pthresh <float> (default 1e-6)\n");
  strcat(usage, "     --bgfile <background> (default from motif file)\n");
  strcat(usage, "     --list <file name>\n");
  strcat(usage, "     --norc (default false)\n");
  strcat(usage, "     --pseudocount <float> (default=0.1)\n");
  strcat(usage, "     --print-col-freqs\n");
  strcat(usage, "     --print-trimmed-tree\n");
  strcat(usage, "     --qvalue\n");
  strcat(usage, "     --gff <filename>\n");
  strcat(usage, "     --ustar <label>\n");	// TLB; create uniform star tree
  strcat(usage, "     --verbosity [1|2|3|4] (default 2)\n");
  strcat(usage, "\n");


  // Parse the command line.
  if (simple_setopt(argc, argv, n_options, motiph_options) != NO_ERROR) {
    die("Error processing command line options: option name too long.\n");
  }

  while (TRUE) { 
    int c = 0;
    char* option_name = NULL;
    char* option_value = NULL;
    const char * message = NULL;

    // Read the next option, and break if we're done.
    c = simple_getopt(&option_name, &option_value, &option_index);
    if (c == 0) {
      break;
    } else if (c < 0) {
      (void) simple_getopterror(&message);
      die("Error processing command line options (%s)\n", message);
    }
    
    if (strcmp(option_name, "model") == 0) {
      if (strcmp(option_value, "jc") == 0) {
        model_type = JC_MODEL;
      } else if (strcmp(option_value, "k2") == 0) {
        model_type = K2_MODEL;
      } else if (strcmp(option_value, "f81") == 0) {
        model_type = F81_MODEL;
      } else if (strcmp(option_value, "f84") == 0) {
        model_type = F84_MODEL;
      } else if (strcmp(option_value, "hky") == 0) {
        model_type = HKY_MODEL;
      } else if (strcmp(option_value, "tn") == 0) {
        model_type = TAMURA_NEI_MODEL;
      } else if (strcmp(option_value, "single") == 0) {
        model_type = SINGLE_MODEL;
      } else if (strcmp(option_value, "average") == 0) {
        model_type = AVERAGE_MODEL;
      } else {
        die("Unknown model: %s\n", option_value);
      }
    } else if (strcmp(option_name, "column-freqs") == 0) {
      if (strcmp(option_value, "simulated") == 0) {
        column_freqs_type = SIMULATED_COL_FREQS;
      } else if (strcmp(option_value, "empirical") == 0) {
        column_freqs_type = EMPIRICAL_COL_FREQS;
      } else {
        die("Unknown column-freqs type: %s\n", option_value);
      }
    } else if (strcmp(option_name, "hb") == 0){
        use_halpern_bruno = TRUE;
    } else if (strcmp(option_name, "ustar") == 0){	// TLB; create uniform star tree
        ustar_label = option_value;
    } else if (strcmp(option_name, "pur-pyr") == 0){
        purine_pyrimidine = atof(option_value);
    } else if (strcmp(option_name, "transition-transversion") == 0){
        transition_transversion = atof(option_value);
    } else if (strcmp(option_name, "bg") == 0){
      bg_rate = atof(option_value);
    } else if (strcmp(option_name, "fg") == 0){
      fg_rate = atof(option_value);
    } else if (strcmp(option_name, "pthresh") == 0){
      pthresh = atof(option_value);
    } else if (strcmp(option_name, "gap") == 0){
      if (strcmp(option_value, "skip") == 0) {
        gap_support = SKIP_GAPS;
      } else if (strcmp(option_value, "fixed") == 0) {
        gap_support = FIXED_GAP_COST;
      } else if (strcmp(option_value, "wildcard") == 0) {
        gap_support = WILDCARD_GAP;
      } else if (strcmp(option_value, "minimum") == 0) {
        gap_support = MIN_GAPS;
      } else {
        die("Unknown gap handling method: %s\n", option_value);
      }
    } else if (strcmp(option_name, "gap-cost") == 0) {
      gap_cost = atof(option_value);
    } else if (strcmp(option_name, "motif") == 0){
        if (selected_motifs == NULL) {
          selected_motifs = new_string_list();
        }
       add_string(option_value, selected_motifs);
    } else if (strcmp(option_name, "motif-name") == 0){
      motif_name = option_value;
    } else if (strcmp(option_name, "list") == 0){
      use_file_list = TRUE;
    } else if (strcmp(option_name, "bgfile") == 0){
      bg_filename = option_value;
    } else if (strcmp(option_name, "sequence-name") == 0){
      seq_name = option_value;
    } else if (strcmp(option_name, "pseudocount") == 0){
      pseudocount = atof(option_value);
    } else if (strcmp(option_name, "norc") == 0) {
      scan_both_strands = FALSE;
    } else if (strcmp(option_name, "print-col-freqs") == 0){
      print_col_freqs = TRUE;
    } else if (strcmp(option_name, "print-trimmed-tree") == 0){
      print_trimmed_tree = TRUE;
    } else if (strcmp(option_name, "qvalue") == 0) {
      compute_qvalues = TRUE;
    } else if (strcmp(option_name, "gff") == 0) {
      gff_filename = option_value;
    } else if (strcmp(option_name, "verbosity") == 0){
      verbosity = atoi(option_value);
    }
  }

  // Must have tree and alignment file names
  if (argc != option_index + 3) {
    fprintf(stderr, usage);
    exit(EXIT_FAILURE);
  } 

  /****************************************************
   * Read the names of the alignments or sequence files
   ****************************************************/
  STRING_LIST_T* filenames = NULL;
  int num_filenames = 0;
  if (use_file_list) {
    filenames = read_string_list_from_file(argv[option_index]);
    num_filenames = get_num_strings(filenames);
  }
  else {
    filenames = new_string_list();
    add_string(argv[option_index], filenames);
    num_filenames = 1;
  }
  option_index++;

  /**********************************************
   * Read the phylogenetic tree.
   **********************************************/
  char* tree_filename = NULL;
  TREE_T* tree = NULL;
  TREE_T* trimmed_tree = NULL;
  tree_filename = argv[option_index];
  option_index++;
  tree = read_tree_from_file(tree_filename);

  /**********************************************
   * Read the motifs.
   **********************************************/
  char* meme_filename = argv[option_index];
  int num_motifs = 0; 
  MOTIF_T motifs[2 * MAX_MOTIFS];
  STRING_LIST_T* motif_occurrences = NULL;
  BOOLEAN_T has_reverse_strand = FALSE;
  ARRAY_T* bg_freqs = NULL;
  option_index++;
  BOOLEAN_T read_file = FALSE;
  // Assume MEME file is in XML format.
  read_file = read_meme_xml_file(
    meme_filename, 
    bg_filename,
    pseudocount,
    &num_motifs, 
    motifs, 
    &motif_occurrences,
    &has_reverse_strand,
    &bg_freqs
  );
  if (!read_file) {
    // MEME file was not in XML format, try old format
    read_meme_file(
      meme_filename, 
      bg_filename,
      pseudocount,
      &num_motifs, 
      motifs, 
      &motif_occurrences,
      &has_reverse_strand, 
      &bg_freqs
    );
  }

  // We don't make use of the motif occurence data
  free_string_list(motif_occurrences);

  if (scan_both_strands == TRUE) {
    // Set up hash tables for computing reverse complement
    setup_hash_alph(DNAB);
    setalph(0);
    // Correct background by averaging on freq. for both strands.
    average_freq_with_complement(bg_freqs);
    int alph_size = get_alph_size(ALPH_SIZE);
    normalize_subarray(0, alph_size, 0.0, bg_freqs);
    fill_in_ambiguous_chars(FALSE, bg_freqs);
    // Make reverse complement motifs.
    int num_motifs_after_rc = num_motifs;
    add_reverse_complements(&num_motifs_after_rc, motifs);
    assert(num_motifs_after_rc == (2 * num_motifs));
    num_motifs = num_motifs_after_rc;
  }

  // TLB; need to resize bg_freqs array to 4 items
  // or copy array breaks in HB mode.  This throws away
  // the freqs for the ambiguous characters;
  // FIXME: fix this so that ambigs are allowed
  resize_array(bg_freqs, 4);

  // Only SIMULATED column frequencies are supported for SINGLE model.
  if (model_type == SINGLE_MODEL && column_freqs_type != SIMULATED_COL_FREQS) {
    if (verbosity >= NORMAL_VERBOSE) {
      fprintf(stderr, "Using simulated column frequencies.\n");
    }
    column_freqs_type = SIMULATED_COL_FREQS;
  }

  // Compute the column frequencies distributions from the input
  // alignments if necessary.
  if (column_freqs_type == EMPIRICAL_COL_FREQS) {
    if (verbosity >= NORMAL_VERBOSE) {
      fprintf(
       stderr, 
       "Computing column frequencies from multiple alignment file(s).\n"
      );
    }
    alignment_col_freqs_list = get_alignment_column_freqs_list(filenames);
  } else if (column_freqs_type == PRECOMPUTED_COL_FREQS) {
    // FIXME: add switch --column-freqs-file <file> and read that
    // file of column frequency distributions for different sets of species.
    die("PRECOMPUTED column frequencies not yet supported.\n");
  } else {
    // Using SIMULATED column frequencies.  Initialize list.
    alignment_col_freqs_list = new_object_list();
  } // column freqs list

  /****************************************************
   * Open the optional GFF file for output.
   ****************************************************/
  FILE* gff_file = NULL;
  if (gff_filename != NULL) {
    if (open_file(gff_filename, "w", TRUE, "GFF", "GFF", &gff_file) == 0) {
      exit(1);
    }
  } else {
    gff_file = stdout;
  }

  /**************************************************************
  * Score each of the alignments for each of the selected motifs.
  **************************************************************/
  int n_scored = 0;    // number of positions that were scored
  int motif_index;
  for (motif_index = 0; motif_index < num_motifs; motif_index++) {

    MOTIF_T* motif = &(motifs[motif_index]);
    char* motif_id = get_motif_id(motif);
    char* bare_motif_id = motif_id;

    // We may have specified on the command line that
    // only certain motifs were to be used.
    if (selected_motifs != NULL) {
      if (*bare_motif_id == '+' || *bare_motif_id == '-') {
        // The selected  motif id won't included a strand indicator.
        bare_motif_id++;
      }
      if (have_string(bare_motif_id, selected_motifs) == FALSE) {
        continue;
      }
    }

    int window_size = motif->length;
    if (verbosity >= NORMAL_VERBOSE) {
      fprintf(
        stderr, 
        "Using motif %s of width %d.\n",
        motif_id, motif->length
      );
    }

    // Build an array of evolutionary models for each position in the motif.
    EVOMODEL_T** models = make_motif_models(
      motif, 
      bg_freqs,
      model_type,
      fg_rate, 
      bg_rate, 
      purine_pyrimidine, 
      transition_transversion, 
      use_halpern_bruno
    );

    STRING_LIST_T* alignment_species = NULL;
    MATRIX_T* pssm = NULL;
    ARRAY_T* alignment_col_freqs = NULL;
    ARRAY_T* pv_lookup = NULL;
    OBJECT_LIST_T* pssm_list = new_object_list();
    OBJECT_LIST_T* pv_lookup_list = new_object_list();
    int file_index;

    SCORED_SITES_T* site_list;
    site_list = new_scored_sites();

    // Consider each alignment in turn.
    for(file_index = 0; file_index < num_filenames; file_index++) { 
      char* filename = get_nth_string(file_index, filenames);

      if (verbosity >= NORMAL_VERBOSE && !(file_index % 100)) {
        fprintf(
          stderr, 
          "Using alignment file number %d of %d total files.\n",
          file_index+1, num_filenames
        );
      }
      ALIGNMENT_T* alignment = read_alignment_from_file(filename, FALSE);

      // Print the tree and then exit?  (Needed for testing Monkey.)
      if (print_trimmed_tree) {
        trimmed_tree = trim_tree_to_alignment(tree, alignment);
        fprintf(stdout, "%s ", filename);
        write_tree(trimmed_tree, stdout);
        free_tree(TRUE, trimmed_tree);
        continue;      // get next alignment
      }

      // Get the names of the species in the alignment.
      alignment_species = get_species_names(alignment);

      // FIXME If the alignment contains species that are not
      // also in the tree memory can be corrupted.
      // Test that every species in the alignment is also in the tree.
      // Die with error message if this is not true.

      // Try to retrieve the pre-computed PSSM from memory.
      pssm = (MATRIX_T*)retrieve_object(alignment_species, pssm_list);

      // Did we fail to retrieve the PSSM?
      if (pssm == NULL) {
        if (verbosity > NORMAL_VERBOSE) {
          fprintf(stderr, "Building PSSM for %s.\n",
          combine_string_list(alignment_species, "-"));
        }

        // Trim the tree to match the species in the alignment.
        trimmed_tree = trim_tree_to_alignment(tree, alignment);

        // TLB; Convert the tree to a uniform star tree with
        // the target sequence at its center.
        if (ustar_label != NULL) {
          trimmed_tree = 
            convert_to_uniform_star_tree(trimmed_tree, ustar_label);
          if (trimmed_tree == NULL) 
            die("Tree or alignment missing target %s\n", ustar_label);
	  if (verbosity >= NORMAL_VERBOSE) {
	    fprintf(stderr, 
              "Target %s placed at center of uniform star tree:\n", 
              ustar_label
            );
	    write_tree(trimmed_tree, stderr);
	  }
        }

        // Build PSSM for motiph
        pssm = build_alignment_pssm(
          alignment_species,
          window_size + 1, 
          models, 
          trimmed_tree, 
          gap_support
        );
        free_tree(TRUE, trimmed_tree);

        // Get the column frequency distribution for computing p-values.
        alignment_col_freqs = 
          retrieve_object(alignment_species, alignment_col_freqs_list);
        // Column frequencies not found?
        if (alignment_col_freqs == NULL) {
          if (column_freqs_type != SIMULATED_COL_FREQS) {
            die("No column frequency distribution was given for species: %s\n",
              combine_string_list(alignment_species, "-"));
          }
          // Get the simulated column frequencies from first row of PSSM.
          // They are based on the single-letter frequencies and the evo model.
          alignment_col_freqs = allocate_array(get_num_cols(pssm)); 
          copy_array(get_matrix_row(0, pssm), alignment_col_freqs);
          // Save the column frequencies for this species list for future use.
          store_object(
            (void*)alignment_col_freqs,
            alignment_species, 
            alignment_col_freqs_list
          );
          if (print_col_freqs && column_freqs_type == SIMULATED_COL_FREQS) {
            print_col_frequencies(alignment_col_freqs);
            exit(1);
          }
        } else if (print_col_freqs) {
          print_col_frequencies(alignment_col_freqs);
          exit(1);
        }

        // Remove first row from the PSSM.  It was saved above if needed.
        remove_matrix_row(0, pssm);

	// Build tables to translate log-odds scores to p-values
	scale_pssm(pssm, RANGE);
	pv_lookup = get_pv_lookup(
				  pssm, 
				  RANGE,
				  get_num_cols(pssm), 
				  alignment_col_freqs
				  );

        // Store them for later use.
        store_object((void*)pssm, alignment_species, pssm_list);
	store_object((void*)pv_lookup, alignment_species,pv_lookup_list);
      }

      // We got the PSSM, so get the other stuff that goes with it.
      else {
        if (verbosity > NORMAL_VERBOSE) {
          fprintf(stderr, "Retrieved PSSM for %s.\n",
          combine_string_list(alignment_species, "-"));
        }

        alignment_col_freqs 
          = (ARRAY_T*) retrieve_object(alignment_species,
        alignment_col_freqs_list);
	pv_lookup = (ARRAY_T*) retrieve_object(
					       alignment_species, 
					       pv_lookup_list
					       );
      }

      // Free list of species
      free_string_list(alignment_species);

      // Build a table for converting the index into the alignment 
      // to an index into ungapped reference sequence.
      int* coord_conv_table = make_alignment_to_seq_table(0, alignment);

      n_scored += score_sequence_in_alignment(
        0, 
        alignment, 
        motif_name, 
        motif->id,
        seq_name, 
        window_size,
        models, 
        pssm,
        pv_lookup,
        coord_conv_table,
        gap_support, 
        gap_cost,
        pthresh,
	site_list
      );
      free_alignment(alignment);
    } // score current alignment

    // Compute q-values, if requested.
    if (compute_qvalues) {
      convert_pvalues_to_qvalues(site_list);
    }

    // Print the sites.
    print_scored_sites(program_name, site_list, gff_file, stdout);
    free_scored_sites(site_list);

    if (models != NULL) {
      int model_index;
      int num_models = motif->length + 1;
      for (model_index = 0; model_index < num_models; model_index++) {
        free_model(models[model_index]);
      }
      myfree(models);
    }
    free_object_list(free_matrix, pssm_list);
    free_object_list(free_array, pv_lookup_list);
  } // score using current motif

  for (motif_index = 0; motif_index < num_motifs; motif_index++) {
    free_matrix(motifs[motif_index].freqs);
  }

  /**********************************************
   * Clean up.
   **********************************************/
  // TLB may have encountered a memory corruption bug here
  // CEG has not been able to reproduce it. valgrind says all is well.
  free_object_list(free_array, alignment_col_freqs_list);
  free_array(bg_freqs);
  free_tree(TRUE, tree);
  free_string_list(filenames);
  free_string_list(selected_motifs);

  //FIXME: (tlb) this is useful for checking p-value accuracy
  fprintf(stderr, "Total number of positions scored: %d\n", n_scored);

  return(0);
} // main

/*******************************************************************
  Print the column frequency distribution.
 ********************************************************************/
static void print_col_frequencies(
  ARRAY_T* alignment_column_freqs
)
{
  int i;
  int num_freqs = get_array_length(alignment_column_freqs);
  int alph_size = get_alph_size(ALPH_SIZE);
  int num_leaves = NINT(log(num_freqs)/log(alph_size));
  char* alignment_col = mm_malloc((num_leaves + 1) * sizeof(char));
  for (i=0; i<num_freqs; i++) {
    unhash_alignment_col(
      i,                              //col_index
      alignment_col,
      num_leaves
    );
    printf("%s %d %g\n", alignment_col, i+1, 
      get_array_item(i, alignment_column_freqs));
  }
} // print_col_freqs


TREE_T* trim_tree_to_alignment(TREE_T* tree, ALIGNMENT_T* alignment) {

  // Trim species from the tree that aren't in the alignment.
  STRING_LIST_T* species_names = get_species_names(alignment);
  TREE_T* trimmed_tree = trim_tree(TRUE, tree, species_names);
  free_string_list(species_names);
  if (verbosity >= HIGH_VERBOSE) {
    fprintf(stderr, "Trimmed tree: ");
    write_tree(trimmed_tree, stderr);
  }

  return trimmed_tree;

}
