/********************************************************************
 * FILE: object-list.c
 * AUTHOR: William Stafford Noble
 * CREATE DATE: May 3, 2006
 * PROJECT: META-MEME
 * COPYRIGHT: 2006, UW
 ********************************************************************/

#include "object-list.h"
#include "string-list.h"


/*************************************************************************
 * Primary data structure.
 *************************************************************************/
struct object_list_t {
  int             num_objects; // Number of objects in the array.
  int             max_objects; // Total amount of memory allocated.
  int             next_object; // Next object to retrieve.
  STRING_LIST_T** keys;        // Each OBJECT is identified by a string-list.
  void**          values;      // The OBJECTs themselves.
};

/*************************************************************************
 * Create an empty object list.
 *************************************************************************/
#define DEFAULT_MAX_OBJECTS 10
OBJECT_LIST_T* new_object_list()
{
  OBJECT_LIST_T* new_object_list;

  new_object_list = (OBJECT_LIST_T*)mm_calloc(1, sizeof(OBJECT_LIST_T));
  new_object_list->num_objects = 0;
  new_object_list->max_objects = DEFAULT_MAX_OBJECTS;
  new_object_list->next_object = 0;
  new_object_list->keys = 
    (STRING_LIST_T**)mm_calloc(DEFAULT_MAX_OBJECTS, sizeof(STRING_LIST_T**));
  new_object_list->values = 
    (void**)mm_calloc(DEFAULT_MAX_OBJECTS, sizeof(void**));

  return(new_object_list);
}

/*************************************************************************
 * Store a object in a list.
 *************************************************************************/
void store_object
 (void*          object,
  STRING_LIST_T* key,
  OBJECT_LIST_T* an_object_list)
{
  // Make sure none of the inputs is null.
  if (object == NULL) {
    die("Adding null OBJECT to list.");
  }
  if (key == NULL) {
    die("Adding OBJECT with null key to list.");
  }
  if (an_object_list == NULL) {
    die("Adding OBJECT to null list.");
  }

  // If it's already there, don't do anything.
  if (retrieve_object(key, an_object_list) != NULL) {
    fprintf(stderr, "Adding OBJECT that was already in the list.\n");
    return;
  }

  // Reallocate space if there isn't any.
  if (an_object_list->num_objects >= an_object_list->max_objects) {
    an_object_list->max_objects += DEFAULT_MAX_OBJECTS;
    an_object_list->keys = (STRING_LIST_T**)mm_realloc(an_object_list->keys,
						   an_object_list->max_objects 
						   * sizeof(STRING_LIST_T*));
    an_object_list->values = (void**)mm_realloc(an_object_list->values,
					      an_object_list->max_objects 
					      * sizeof(void*));
  }

  // Add to the list.
  //an_object_list->keys[an_object_list->num_objects] = key;
  an_object_list->keys[an_object_list->num_objects] = copy_string_list(key);
  an_object_list->values[an_object_list->num_objects] = object;
  (an_object_list->num_objects)++;
  if (verbosity > NORMAL_VERBOSE) {
    fprintf(
	  stderr,
	  "Added OBJECT %d to list with key %s.\n",
	  an_object_list->num_objects,
	  combine_string_list(an_object_list->keys[(an_object_list->num_objects) - 1], "-"));
  }
}

/*************************************************************************
 * Retrieve a object from a list.
 *************************************************************************/
void* retrieve_object
 (STRING_LIST_T* key,
  OBJECT_LIST_T* an_object_list)
{
  int i_object;

  for (i_object = 0; i_object < an_object_list->num_objects; i_object++) {
    if (equal_string_lists(key, an_object_list->keys[i_object])) {
      return(an_object_list->values[i_object]);
    }
  }
  // If we didn't find it, return null pointer.
  return(NULL);
}

/*************************************************************************
 * Get the next object in a list; saves state.  
 * First call retrieves first object; successive calls retrieve next ones.
 * Returns NULL if at end * of list.  
 * The next call after that will retrieve the first object again.
 *************************************************************************/
void* retrieve_next_object
 (OBJECT_LIST_T* an_object_list)
{
  if (an_object_list->next_object == an_object_list->num_objects) {
    an_object_list->next_object = 0;		// rewind
    return(NULL);
  }
  else {
    return(an_object_list->values[an_object_list->next_object++]);
  }
}

/*************************************************************************
 * Free dynamic memory used by a list of objects.
 *************************************************************************/
void free_object_list
(void (*free_object)(),
 OBJECT_LIST_T* a_list)
{
  int i_object;

  if (a_list == NULL) {
    return;
  }

  for (i_object = 0; i_object < a_list->num_objects; i_object++) {
    free_string_list(a_list->keys[i_object]);
    (free_object)(a_list->values[i_object]);
  }
  myfree(a_list->keys);
  myfree(a_list->values);
  myfree(a_list);
}

/*
 * Local Variables:
 * mode: c++
 * c-basic-offset: 2
 * End:
 */
