/**************************************************************************
 * FILE: tomtom.c
 * AUTHOR: Shobhit Gupta and Timothy Bailey
 * CREATE DATE: 02-21-06
 * PROJECT: TOMTOM
 * DESCRIPTION: Motif-Motif comparison using the score statistic.
 **************************************************************************/
#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include "utils.h"       // Generic utilities.
#include "motif.h"       // Needed for complement_dna_freqs.
#include "mhmm-state.h"  // HMM.
#include "matrix.h"      // Routines for floating point matrices.
#include "array.h"       // Routines for floating point arrays.
#include "alphabet.h"    // The alphabet.
#include "fasta-io.h"
#include "meme-io.h"
#include "simple-getopt.h"
#include "pssm.h"
#include "tomtom.h"
#include "fitevd.h"
#include "float.h"
#include "config.h"
#include "io.h"
#include "ceqlogo.h"

// URL formats for XML output; update if these on-line services move
#define JASPAR_URL_FORMAT \
  "<a href='http://jaspar.genereg.net/cgi-bin/jaspar_db.pl?rm=present&db=CORE&ID=%s'>%s</a>"
#define TRANSFAC_URL_FORMAT \
  "<a href='http://www.biobase.de/cgi-bin/biobase/transfac/11.1/bin/getTFProf.cgi?%s'>%s</a>"
#define SCPD_URL_FORMAT \
  "<a href='http://rulai.cshl.edu/cgi-bin/SCPD/getfactor?%s'>%s</a>"
#define MACISAAC_URL_FORMAT \
  "<a href='http://fraenkel.mit.edu/improved_map/v1.tamo'>%s</a>"
#define FLYREG_URL_FORMAT \
  "<a href='http://www.danielpollard.com/bergman2004_matrices.html'>%s</a>"
//  "<a href='http://www.flyreg.org'>%s</a>"
#define DPINTERACT_URL_FORMAT \
  "<a href='http://arep.med.harvard.edu/ecoli_matrices/%s.html'>%s</a>"
#define REGTRANSBASE_URL_FORMAT \
  "<a href='http://regtransbase.lbl.gov/cgi-bin/regtransbase?page=alignment_browse'>%s</a>"
#define PRODORIC_URL_FORMAT \
  "<a href='http://prodoric.tu-bs.de/matrix.php?matrix_acc=%s'>%s</a>"

#define DEFAULT_QUERY_PSEUDO   0.0
#define DEFAULT_TARGET_PSEUDO   0.0
#define MAX_TARGETS 5000
#define MAX_QUERYS 5000
#define BINS 100
#define ITERATIONS 1000
#define LOGOHEIGHT 10		// height of an alignment of logos in cm.
#define MIN_TARGET_DATABASE_SIZE 50

/* Define some string limits */
// FIXME: these are brittle and will cause bugs if motifs are too long
int MAX_MOTIF_LENGTH = 100;
int MAX_LINE_LENGTH = 100;

VERBOSE_T verbosity = NORMAL_VERBOSE;

#ifdef MAIN

// Local functions
void print_tomtom_xml_header(
  FILE *output_path,
  char *query_filename,
  char *target_filename,
  char *distance_measure,
  float ethresh
); 
char *create_logo(
  MOTIF_T* target,             // Target motif
  MOTIF_T* query,              // Query motif
  int offset,                  // Offset of target rel. to query
  char *output_dirname 	       // Output directory
);
void print_tomtom_xml_line(
  FILE *output_path,		// path to output XML file
  char *line,			// stored results line
  char *output_dirname,		// name of output directory
  MOTIF_T *motifs_query,
  MOTIF_T *motifs_target,
  char *query_url_format,	// format string for query URL
  char *target_url_format 	// format string for target URL
);
void print_tomtom_xml_tailer(
  FILE *output_path,
  int argc,
  char **argv,
  int total_matches
); 


int main(int argc, char *argv[])
{
  char *default_output_dirname = "tomtom_out"; // where to write HTML
  char *text_filename = "tomtom.txt";
  FILE *text_output;			// where to write legacy output
  char *xml_filename = "tomtom.html";	// FIXME: change to .xml and output XML
  FILE *xml_output = NULL;
  //char *html_filename = "tomtom.html";// FIXME: convert XML to HTML
  //FILE *html_output = NULL;

  /**********************************************
   * Initialize command line parameters
   **********************************************/
  char* query_filename = NULL;
  char* target_filename = NULL;
  char *output_dirname = default_output_dirname;
  BOOLEAN_T clobber = FALSE;		// don't clobber existing files
  BOOLEAN_T text_only = FALSE; 		// default is HTML
  char* distance_measure = "pearson";
  float ethresh = 10;
  char* program_mode = "normal";
  char *target_url_format = "%s"; 	// no URL format
  char *query_url_format = "%s";	// no URL format
  float query_pseudocount = DEFAULT_QUERY_PSEUDO; 
  float target_pseudocount = DEFAULT_TARGET_PSEUDO; 


  /**********************************************
   * Command line parsing.
   **********************************************/
  int option_count = 14;     // Must be updated if list below is
  cmdoption const motif_scan_options[] = {
    {"query", REQUIRED_VALUE},
    {"target", REQUIRED_VALUE},
    {"o", OPTIONAL_VALUE},
    {"oc", OPTIONAL_VALUE},
    {"text", OPTIONAL_VALUE},
    {"ethresh", OPTIONAL_VALUE},
    {"dist", OPTIONAL_VALUE},
    {"mode", OPTIONAL_VALUE},
    {"internal", OPTIONAL_VALUE},
    {"min-overlap", OPTIONAL_VALUE},
    {"query-pseudo", OPTIONAL_VALUE},
    {"target-pseudo", OPTIONAL_VALUE},
    {"query-url-type", OPTIONAL_VALUE},
    {"target-url-type", OPTIONAL_VALUE}
  };

  int option_index = 0;

   /**********************************************
    * Define the usage message.
    **********************************************/
  char usage[2000] = "";
  strcat(usage, "\n   USAGE: tomtom [options] -query <query file> -target <target file>\n");
  strcat(usage, "\n");
  strcat(usage, "   Options:\n");
  strcat(usage, "     -o <output dir>\tname of directory for output files;\n\t\t\t\twill not replace existing directory\n");
  strcat(usage, "     -oc <output dir>\tname of directory for output files;\n\t\t\t\twill replace existing directory\n");
  strcat(usage, "     -ethresh <float>\t\tE-value cutoff for alignments; default: 10\n");
  strcat(usage, "     -dist allr|ed|kullback|pearson|sandelin\n\t\t\t\tdistance metric for scoring alignments;\n\t\t\t\tdefault: pearson (Pearson correlation coefficient)\n");
  strcat(usage, "     -internal\t\t\tonly allow internal alignments;\n\t\t\t\tdefault: allow overhangs\n");
  strcat(usage, "     -min-overlap <int>\t\tminimum overlap between query and target;\n\t\t\t\tdefault: 1\n");
  strcat(usage, "     -query-pseudo <float>\tdefault: %.1f\n");
  strcat(usage, "     -target-pseudo <float>\tdefault: %.1f\n");
  strcat(usage, "     -text\t\t\toutput in text format (default is HTML)\n");
  strcat(usage, "     -query-url-type jaspar|transfac|scpd|macisaac|flyreg|dpinteract|regtransbase|none\n\t\t\t\tlink query motifs to this on-line DB;\n\t\t\t\tdefault: none\n");
  strcat(usage, "     -target-url-type jaspar|transfac|scpd|macisaac|flyreg|dpinteract|regtransbase|none\n\t\t\t\tlink target motifs to this on-line DB;\n\t\t\t\tdefault: none\n");

  strcat(usage, "\n");

   /**********************************************
    * Parse the command line.
    **********************************************/
   if (simple_setopt(argc, argv, option_count, motif_scan_options) != NO_ERROR)
     {
       die("Error: option name too long.\n");
     }
   BOOLEAN_T internal = FALSE;
   int min_overlap = 1;
   while (TRUE) {
    int c = 0;
    char* option_name = NULL;
    char* option_value = NULL;
    const char * message = NULL;

    /*  Read the next option, and break if we're done. */
    c = simple_getopt(&option_name, &option_value, &option_index);
    if (c == 0) {
      break;
    } else if (c < 0) {
      (void) simple_getopterror(&message);
      die("Error processing command line options (%s)\n", message);
    }

    if (strcmp(option_name, "ethresh") == 0) {
      if (! option_value) {
        fprintf(stderr, "error: missing value after -ethresh.\n");
        exit(EXIT_FAILURE);
      }
      ethresh = atof(option_value);
    } else if (strcmp(option_name, "query") == 0) {
      query_filename = option_value;
    } else if (strcmp(option_name, "target") == 0) {
      target_filename = option_value;
    } else if (strcmp(option_name, "o") == 0) {
      output_dirname = option_value;
    } else if (strcmp(option_name, "oc") == 0) {
      output_dirname = option_value;
      clobber = TRUE;		// OK to clobber
    } else if (strcmp(option_name, "text") == 0) {
      text_only = TRUE;
    } else if (strcmp(option_name, "dist") == 0) {
      distance_measure = option_value;
    } else if (strcmp(option_name, "mode") == 0) {
      program_mode = option_value;
    } else if (strcmp(option_name, "internal") == 0) {
      internal = TRUE;
    } else if (strcmp(option_name, "min-overlap") == 0) {
      double temp_var;
      if (! option_value) {
        fprintf(stderr, "error: missing value after -min-overlap.\n");
        exit(EXIT_FAILURE);
      }
      temp_var = atof(option_value);
      min_overlap = rint(temp_var);
    } else if (strcmp(option_name, "query-pseudo") == 0) {
      if (! option_value) {
        fprintf(stderr, "error: missing value after -query-pseudo.\n");
        exit(EXIT_FAILURE);
      }
      query_pseudocount = atof(option_value);
    } else if (strcmp(option_name, "target-pseudo") == 0) {
      if (! option_value) {
        fprintf(stderr, "error: missing value after -target-pseudo.\n");
        exit(EXIT_FAILURE);
      }
      target_pseudocount = atof(option_value);
    } else if (strcmp(option_name, "query-url-type") == 0) {
       if (strcmp(option_value, "jaspar") == 0) {
         query_url_format = JASPAR_URL_FORMAT;
       } else if (strcmp(option_value, "transfac") == 0) {
         query_url_format = TRANSFAC_URL_FORMAT;
       } else if (strcmp(option_value, "scpd") == 0) {
         query_url_format = SCPD_URL_FORMAT;
       } else if (strcmp(option_value, "macisaac") == 0) {
         query_url_format = MACISAAC_URL_FORMAT;
       } else if (strcmp(option_value, "flyreg") == 0) {
         query_url_format = FLYREG_URL_FORMAT;
       } else if (strcmp(option_value, "dpinteract") == 0) {
         query_url_format = DPINTERACT_URL_FORMAT;
       } else if (strcmp(option_value, "regtransbase") == 0) {
         query_url_format = REGTRANSBASE_URL_FORMAT;
       } else if (strcmp(option_value, "prodoric") == 0) {
         query_url_format = PRODORIC_URL_FORMAT;
       } else {
         query_url_format = "%s";
       }
    } else if (strcmp(option_name, "target-url-type") == 0) {
       if (strcmp(option_value, "jaspar") == 0) {
         target_url_format = JASPAR_URL_FORMAT;
       } else if (strcmp(option_value, "transfac") == 0) {
         target_url_format = TRANSFAC_URL_FORMAT;
       } else if (strcmp(option_value, "scpd") == 0) {
         target_url_format = SCPD_URL_FORMAT;
       } else if (strcmp(option_value, "macisaac") == 0) {
         target_url_format = MACISAAC_URL_FORMAT;
       } else if (strcmp(option_value, "flyreg") == 0) {
         target_url_format = FLYREG_URL_FORMAT;
       } else if (strcmp(option_value, "dpinteract") == 0) {
         target_url_format = DPINTERACT_URL_FORMAT;
       } else if (strcmp(option_value, "regtransbase") == 0) {
         target_url_format = REGTRANSBASE_URL_FORMAT;
       } else if (strcmp(option_value, "prodoric") == 0) {
         target_url_format = PRODORIC_URL_FORMAT;
       } else {
         target_url_format = "%s";
       }
    }
  }
  if ((query_filename == NULL)||(target_filename == NULL)) {
    fprintf(stderr, usage, DEFAULT_QUERY_PSEUDO, DEFAULT_QUERY_PSEUDO);
    exit(EXIT_FAILURE);
  }
  if (text_only == TRUE) {
    // Legacy: plain text output to standard out.
    text_output = stdout;
  } else {
    // allow clobbering of the default output directory
    if (output_dirname == default_output_dirname) {
      clobber = TRUE;
    }
    if (create_output_directory(output_dirname, clobber, TRUE)) {
      // Failed to create output directory.       
      die("Unable to create output directory %s.\n", output_dirname);
    }
    // Create the name of the output files (text, XML, HTML) and open them
    char *path;
    path = make_path_to_file(output_dirname, text_filename);
    text_output = fopen(path, "w"); //FIXME CEG check for errors
    myfree(path);
    path = make_path_to_file(output_dirname, xml_filename);
    xml_output = fopen(path, "w"); //FIXME CEG check for errors
    myfree(path);
    //path = make_path_to_file(output_dirname, html_filename);
    //html_output = fopen(path, "w"); //FIXME CEG check for errors
    //myfree(path);
  }

  /**********************************************
   * Read the motifs.
   **********************************************/

  /* General variables for reading the meme file */
  int num_motifs_query = 0;
  BOOLEAN_T has_reverse_strand_query = FALSE;
  MOTIF_T motifs_query[MAX_QUERYS];
  STRING_LIST_T* motif_occurrences = NULL;
  ARRAY_T* bg_freqs_query;
  read_meme_file(
		query_filename,
    "motif-file", // background freq. filename
    query_pseudocount,
		&num_motifs_query,
		motifs_query,
		&motif_occurrences,
		&has_reverse_strand_query,
		&bg_freqs_query
	);

  int num_motifs_target = 0;
  MOTIF_T motifs_target[MAX_TARGETS];
  ARRAY_T* bg_freqs_target = NULL;
  BOOLEAN_T has_reverse_strand_target = FALSE;
  read_meme_file(
		target_filename,
    "motif-file", // background freq. filename
    target_pseudocount,
		&num_motifs_target,
		motifs_target,
		&motif_occurrences,
		&has_reverse_strand_target,
		&bg_freqs_target
	);
  add_reverse_complements(&num_motifs_target, motifs_target);


  /* Check for Small target databases (multiply by 2 for reverse complements) */
  if (num_motifs_target < (MIN_TARGET_DATABASE_SIZE * 2)) {
    fprintf(stderr,
      "\nWarning: Target database size too small (%d). Provide at least %d motifs for accurate p-value computation\n\n",
      num_motifs_target/2, MIN_TARGET_DATABASE_SIZE);
  }
  int motif_index_query;
  int motif_index_target;
  int total_target_length = 0;
  int max_motif_length = 0;

  for(motif_index_target = 0; motif_index_target < num_motifs_target;
      motif_index_target++){
    MOTIF_T* motif_target = &(motifs_target[motif_index_target]);
    total_target_length += motif_target->length;
    if (max_motif_length < motif_target->length){
      max_motif_length = motif_target->length;
    }
    /*   Remove the ambiguous letter columns */
    while(get_num_cols(motif_target->freqs) > motif_target->alph_size)
      {
  remove_matrix_col(motif_target->alph_size, motif_target->freqs);
      }
  }

  /*   Remove the ambiguous letter columns for query*/
  for(motif_index_query = 0; motif_index_query < num_motifs_query;
      motif_index_query++){
    MOTIF_T* motif_query = &(motifs_query[motif_index_query]);
    while(get_num_cols(motif_query->freqs) > motif_query->alph_size)
      {
  remove_matrix_col(motif_query->alph_size, motif_query->freqs);
      }
    if (max_motif_length < motif_query->length){
      max_motif_length = motif_query->length;
    }
  }

  /* Background distribution array for the infinite alphabet*/
  ARRAY_T* bg_freqs_mod;
  bg_freqs_mod = allocate_array(total_target_length);
  ATYPE bf;
  bf = 1 / ( (double) total_target_length);
  init_array(bf, bg_freqs_mod);

  /*  Recurse over the set of query motifs and target motifs */
  int total_matches = 0;

  // Print output
  fprintf(text_output, "#Query ID\tTarget ID\tOptimal offset");
  fprintf(text_output, "\tE-value\tOverlapping bases\t");
  fprintf(text_output, "Query consensus\tTarget consensus\tOrientation\n");
  if (! text_only) { 
    print_tomtom_xml_header(
      xml_output, 
      query_filename, 
      target_filename,
      distance_measure,
      ethresh
    ); 
  }

  /* Decide the number of Runs based on the mode */
  int query_iterations = num_motifs_query;
  int num_motifs_database = num_motifs_target;
  if (strcmp(program_mode, "dep") == 0) {
    num_motifs_database = num_motifs_target - 1;
  } else if (strcmp(program_mode, "qq") == 0) {
    query_iterations = ITERATIONS;
    num_motifs_database = num_motifs_target - 1;
  }

  /* Recurse over the set of queries. */
  for(motif_index_query = 0; motif_index_query < query_iterations;
      motif_index_query++){
    STRING_LIST_T * printlines_list;
    printlines_list = new_string_list();

    fprintf(stderr, "Processing query %d out of %d \n",
      motif_index_query + 1, query_iterations);
    int curr_query_index = motif_index_query;
    int target_database_length = total_target_length;
    int num_motifs_database = num_motifs_target;

    /* Take a random query index if NOT in normal mode */
    if (strcmp(program_mode, "normal") != 0) {
      curr_query_index = my_drand() * (num_motifs_query-1);
    }

    MOTIF_T* motif_query = &(motifs_query[curr_query_index]);

    if (strcmp(program_mode, "normal") != 0) {
      target_database_length = total_target_length -
        motif_query->length;
      num_motifs_database = num_motifs_target - 1;
    }

    /*  Initialize an array to store the columnwise scores. */
    MATRIX_T* all_columnwise_scores;
    all_columnwise_scores = allocate_matrix(
              motif_query->length,
              target_database_length
              );

    int index_col_all_columnwise_scores = 0;
    int *target_lengths; /* Array to store lengths of target motifs */
    target_lengths = malloc(num_motifs_database * sizeof(int));
    int target_lengths_index = 0;
    STRING_LIST_T* motif_id_target = new_string_list();
    STRING_LIST_T* motif_cons_target = new_string_list();

    /* Recurse over the set of targets. */
    for(motif_index_target = 0; motif_index_target < num_motifs_target;
  motif_index_target++){
      MOTIF_T* motif_target = &(motifs_target[motif_index_target]);  ;

      /* Allow Query & target to be same only in normal mode */
      BOOLEAN_T evaluate = TRUE;
      if (strcmp(program_mode, "normal") != 0) {
  char q_id[100];
  strcpy(q_id, "+");
  strcat(q_id, get_motif_id(motif_query));
  if (strcmp(get_motif_id(motif_target),
       q_id) == 0) {
    evaluate = FALSE;
  }
      }

      if (evaluate) {
  add_string(get_motif_id(motif_target), motif_id_target);

  /* Add target Consensus Sequence to string list*/
  char* target_cons;
  get_cons(motif_target->freqs, &target_cons);
  add_string(target_cons, motif_cons_target);

  /* Compute columnwise scores */
  MATRIX_T* pairwise_column_scores;
  pairwise_column_scores = allocate_matrix(
             motif_query->length,
             motif_target->length
             );

  /* Compute score matrix based on the distance measure */
  if (strcmp(distance_measure, "pearson") == 0) {
    pearson_scores(
       motif_query,
       motif_target,
       bg_freqs_target,
       &pairwise_column_scores,
       program_mode
       );
  } else if (strcmp(distance_measure, "allr") == 0) {
    allr_scores(
          motif_query,
          motif_target,
          bg_freqs_target,
          &pairwise_column_scores,
          program_mode
          );
  } else if (strcmp(distance_measure, "ed") == 0) {
    ed_scores(
         motif_query,
         motif_target,
         bg_freqs_target,
         &pairwise_column_scores,
         program_mode
         );
  } else if (strcmp(distance_measure, "kullback") == 0) {
    kullback_scores(
        motif_query,
        motif_target,
        bg_freqs_target,
        &pairwise_column_scores,
        program_mode
        );
  } else if (strcmp(distance_measure, "sandelin") == 0) {
    sandelin_scores(
        motif_query,
        motif_target,
        bg_freqs_target,
        &pairwise_column_scores,
        program_mode
        );
  } else {
    fprintf(stderr,
      "Invalid Distance measure [allr|ed|kullback|pearson|sandelin]\n"
      );
  }

  /* Update the background frequency array */
  int index_rows;
  for(index_rows = 0;index_rows < pairwise_column_scores->num_rows;
      index_rows++){
    int index_columns;
    for(index_columns = 0;
        index_columns < pairwise_column_scores->num_cols;
        index_columns++){
      set_matrix_cell(
          index_rows,
          index_columns + index_col_all_columnwise_scores,
          get_matrix_cell(index_rows,
              index_columns,
              pairwise_column_scores),
          all_columnwise_scores
          );
    }
  }

  index_col_all_columnwise_scores +=
    get_num_cols(pairwise_column_scores);
  target_lengths[target_lengths_index] = motif_target->length;
  target_lengths_index++;
  free_matrix(pairwise_column_scores);

      } /* The check for same query target */
    }/* Parsing all targets for the query done. */

    /* Scale the pssm */
    scale_pssm(all_columnwise_scores, BINS);

      /* Shuffle database if not in normal mode */
    if (strcmp(program_mode, "normal") != 0) {

      /* Shuffle the target database columns */
      shuffle_matrix_cols(all_columnwise_scores, FALSE);

      /* Shuffle the Query */
      shuffle_matrix_cols(all_columnwise_scores, TRUE);
    }

    /* Compute the pvalue distributions */
    int num_pv_lookup_array;
    num_pv_lookup_array = (pow(all_columnwise_scores->num_rows, 2)
         + (all_columnwise_scores->num_rows))/2;

    int address_index = 0;
    MATRIX_T* reference_matrix;
    reference_matrix = allocate_matrix(
               all_columnwise_scores->num_rows,
               all_columnwise_scores->num_rows
               );
    init_matrix(-1, reference_matrix);
    MATRIX_T* pv_lookup_matrix;

    int num_array_lookups = (BINS * max_motif_length) + 1;
    pv_lookup_matrix = allocate_matrix(
               num_pv_lookup_array,
               num_array_lookups
               );
    init_matrix(0, pv_lookup_matrix);

    /* Compute the motif scores */
    int colid = 0; /* Stores the lengths  of already read columns */
    for(motif_index_target = 0;
      motif_index_target < num_motifs_database; motif_index_target++) {

      MATRIX_T* pairwise_subset;
      pairwise_subset = allocate_matrix(
          motif_query->length,
          target_lengths[motif_index_target]
          );

      /* Get a query-target pair from the overall score matrix. */
      int col_index;
      for(col_index = 0; col_index < target_lengths[motif_index_target];
        col_index++){
	ARRAY_T* temp_array = NULL;
	temp_array = get_matrix_column(col_index + colid,
          all_columnwise_scores);
	set_matrix_column(temp_array, col_index, pairwise_subset);
	free_array(temp_array);
      }

      /* Compute the score and offset wrt the minimum pvalue CORRECTED. */
      double ALLR_SCORE = 0;
      int optimal_offset = -1;
      double optimal_pvalue = 1;
      int overlapping_bases = -1;

      ALLR_SCORE = compare_motifs(
          pairwise_subset,
          &optimal_offset,
          reference_matrix,
          pv_lookup_matrix,
          &optimal_pvalue,
          program_mode,
          &overlapping_bases,
          distance_measure,
          internal,
          min_overlap,
          &address_index,
          all_columnwise_scores,
          target_database_length,
          bg_freqs_mod
          );

      /* Override min-overlap if query is smaller than min-overlap */
      int mo;
      mo = min_overlap;
      if (pairwise_subset->num_rows < mo) {
        mo = pairwise_subset->num_rows;
      }
      if (pairwise_subset->num_cols < mo) {
        mo = pairwise_subset->num_cols;
      }
      /* Correct the motif pvalue (minimization across several pvalues).
       Configurations mulitiplied by 2, because later the minimum of
       the two strands will be taken LATER. Both stands have the same
       number of total_configurations, thus muliplication by 2.*/
      int total_configurations = pairwise_subset->num_rows
        + pairwise_subset->num_cols - (2 * mo) + 1;

      /* Correct the total number of configurations for internal alignments */
      if (internal) {
        total_configurations = abs(pairwise_subset->num_rows
          - pairwise_subset->num_cols) + 1;
      }
      double motif_pvalue = EV(optimal_pvalue, total_configurations * 2);

      /* Correct for multiple testing (evalue) */
      double evalue = motif_pvalue * num_motifs_target / 2.0;

      /* Check for pvalue and evalue cutoffs (if specified) */
      /* -1 means that the threshold was not defined */
      int print_comp = 1;
      if ((ethresh != -1) && (evalue > ethresh)){
        print_comp = 0;
      }

      /* Consensus Sequence */
      char* query_cons;
      get_cons(motif_query->freqs, &query_cons);

      /* Offset_by_length */
      int offset_by_length = optimal_offset - motif_query->length;

      /* Print out the optimal overlap of query with target */
      if (print_comp == 1) {
        char pline[1000];
	/* Part of the string before | serves as storage for the motif
	   indicies and the offset. Hey, I know it's dirty but what else
	   can be done elegantly in this setup */
	sprintf(pline, "%d,%d,%d|%s\t%s\t%d\t%g\t%d\t%s\t%s\t%g",
	  curr_query_index, motif_index_target, offset_by_length,
	  motif_query->id, get_nth_string(motif_index_target, motif_id_target),
	  offset_by_length, evalue,
	  overlapping_bases, query_cons,
	  get_nth_string(motif_index_target, motif_cons_target),
	  //optimal_pvalue);
	  motif_pvalue);
	add_string_with_score(pline, printlines_list, evalue);

	// ensures that create_logo works
	assert(motif_query == &(motifs_query[curr_query_index]));
      }

      colid += target_lengths[motif_index_target];
      free_matrix(pairwise_subset);
    } /* Scores computed for all targets */

    /* Sort the stringlist */
    sort_string_list_by_score(printlines_list, FALSE);

    /* Free all the matrices/arrays related to the current query */
    free_matrix(reference_matrix);
    free_matrix(pv_lookup_matrix);
    free_matrix(all_columnwise_scores);
    free_string_list(motif_id_target);
    free_string_list(motif_cons_target);
    free(target_lengths);

    STRING_LIST_T* printlines_final;
    printlines_final = new_string_list();
    select_strand(printlines_list, printlines_final);

    int i_stored;
    total_matches += get_num_strings(printlines_final);
    for (i_stored = 0; i_stored < get_num_strings(printlines_final);
      i_stored++) {
      char *line = get_nth_string(i_stored, printlines_final);

      // Print text output
      fprintf(text_output, 
        "%s\n", strchr(line, '|')+1);  // skip storage part of line

      // Print XML output
      if (! text_only) {
        // print one alignment
        print_tomtom_xml_line(
          xml_output, 
          line, 
          output_dirname,
          motifs_query,
          motifs_target,
          query_url_format,
          target_url_format
        );
      }
    }

    free_string_list(printlines_final);
    free_string_list(printlines_list);

 }/*  The loop across the queries. */

  // print XML tailer
  if (! text_only) {
    print_tomtom_xml_tailer(xml_output, argc, argv, total_matches);
  }

  free_motif(motifs_query);
  free_motif(motifs_target);
  free_array(bg_freqs_query);
  free_array(bg_freqs_target);
  free_array(bg_freqs_mod);
  return (0);
}/* Main tomtom*/
#endif

/**************************************************************************
*
* Creates a logo of the aligned target and query motif, writes it to an
* EPS file and calls convert to convert the EPS file to a PNG image.
*
* Returns name of logo file (minus the extension).
*
**************************************************************************/
char *create_logo(
  MOTIF_T* target,             // Target motif
  MOTIF_T* query,              // Query motif
  int offset,                  // Offset of target rel. to query
  char *output_dirname 	       // Output directory
) 
{
  char *logo_filename, *logo_path;
  double logo_height = LOGOHEIGHT;
  double logo_width = offset <= 0 ? MAX(query->length, target->length)-offset : 
    MAX(query->length+offset, target->length);

  // create name of logo file
  logo_filename = (char*) malloc(
    sizeof(char) * (5+strlen(query->id)+1+strlen(target->id)+1)
  );
  sprintf(logo_filename, "logo_%s_%s", query->id, target->id);

  // create output file path
  logo_path = make_path_to_file(output_dirname, logo_filename);

  CL_create2(
    target, 		// first motif
    target->id+1,	// title (skip +/- character)
    query, 		// second motif
    query->id, 		// title
    TRUE,               // error bars
    TRUE,               // small sample correction
    logo_height,        // logo height (cm)
    logo_width,         // logo width (cm)
    get_alphabet(FALSE),// alphabet
    -offset, 		// offset of second motif
    logo_path,		// output file path
    "Tomtom"		// program name
  );
  myfree(logo_path);

  return(logo_filename);
}


/**************************************************************************
*
* get_pv_lookup_new
*
* Create a lookup table for motif scores.
*
* Returns an array of pdf values:
*   pv[x] = Pr(score >= x)
* for 0<=x<=range*w.
*
* Assumes:
*   1) motif scores are non-negative, integral
*   2) background model is 0-order Markov
*
* The last four parameters are specific to tomtom (described below).
* For all other usege, these frour parameters can be passed as
* NULL, -1,  NULL, NULL.
*
* Four additional parameters for tomtom:
* Four additional parameters are added for tomtom to store the
* intermediate arrays. For this purpose, a pvalue matrix is passed
* on together with a reference matrix. One integer pointer is passed
* that stores the last row number of the p-value matrix. Another integer
* is passed to report which query column to start with
*
**************************************************************************/
ARRAY_T* get_pv_lookup_new(
MATRIX_T* pssm,     // The PSSM.
int range,          // Maximum score in PSSM.
int alen,           // Alphabet size.
ARRAY_T* background,  // The background model.
int index_start, //Start position of query motif column (tomtom only)
int * address_index,  // index: last row of pv_lookup_matrix (tomtom only)
MATRIX_T* reference_matrix, // reference_matrix (tomtom only)
MATRIX_T* pv_lookup_matrix // pv_lookup_matrix (tomtom only)
) {
int i;
int w = get_num_rows(pssm);
int size = w*range+1;
ARRAY_T* pv;

/* Get the pdf. */
pv = get_pdf_lookup_new(pssm, w, alen, range, background,
      index_start, &(*address_index),
      reference_matrix, pv_lookup_matrix);

/* Get 1-cdf from the pdf. */
for (i=size-2; i>=0; i--) {
  double p = get_array_item(i, pv) + get_array_item(i+1, pv);
  set_array_item(i, MIN(1.0, p), pv);
}

/* Return pvalue lookup table. */
return(pv);
} // get_pv_lookup


/**************************************************************************
*
* get_pdf_lookup_new
*
* Create a lookup table for the pdf.
*
* Returns an array of pdf values:
*   pdf[x] = Pr(score == x)
* for 0<=x<=range*w.
*
* Assumes:
*   1) motif scores are non-negative, integral
*   2) background model is 0-order Markov
*
* The last four parameters are specific to tomtom (described below).
* For all other usege, these frour parameters can be passed as
* NULL, -1,  NULL, NULL.
*
* Four additional parameters for tomtom:
* Four additional parameters are added for tomtom to store the
* intermediate arrays. For this purpose, a pvalue matrix is passed
* on together with a reference matrix. One integer pointer is passed
* that stores the last row number of the p-value matrix. Another integer
* is passed to report which query column to start with
*
**************************************************************************/
static ARRAY_T* get_pdf_lookup_new(
MATRIX_T* pssm,   // The PSSM.
int w,      // Width of PSSM.
int alen,     // Length of alphabet.
int range,    // Range of scores.
ARRAY_T* background,  // Background model.
int index_start,      //Start position of query motif column (tomtom only)
int * address_index, // index: last row of pv_lookup_matrix (tomtom only)
MATRIX_T* reference_matrix, // reference_matrix (tomtom only)
MATRIX_T* pv_lookup_matrix // pv_lookup_matrix (tomtom only)
)
{
int start = index_start;
int stop = index_start;

int i, j, k;
int size = w*range+1;
ARRAY_T* pdf_old = allocate_array(size);
ARRAY_T* pdf_new = allocate_array(size);

init_array(0, pdf_new);
set_array_item(0, 1, pdf_new);    // Prob(0)

// Compute the pdf recursively.
for (i=0; i<w; i++) {
  int max = i * range;
  SWAP(ARRAY_T*, pdf_new, pdf_old)
  for (k=0; k<=max+range; k++) set_array_item(k, 0, pdf_new);
  for (j=0; j<alen; j++) {
    int s = (int) get_matrix_cell(i, j, pssm);
    for(k=0; k<=max; k++) {
      double old = get_array_item(k, pdf_old);
      if (old != 0) {
	double new = get_array_item(k+s, pdf_new) +
	  (old * get_array_item(j, background));
	set_array_item(k+s, new, pdf_new);
      } // old
    } // k
  } // j

  if (reference_matrix != NULL) {
    int curr_address_index = 0;
    curr_address_index = (int) get_matrix_cell(start,
	   stop,
	   reference_matrix);
    if (curr_address_index == -1) {
ARRAY_T* transient_array;
transient_array = allocate_array(get_array_length(pdf_new));
copy_array(pdf_new, transient_array);

/* Get 1-cdf from the pdf. */
int ii;
for (ii=size-2; ii>=0; ii--) {
  double p = get_array_item(ii, transient_array) +
    get_array_item(ii+1, transient_array);
  set_array_item(ii, MIN(1.0, p), transient_array);
}

int temp_index;
for(temp_index = 0; temp_index < get_array_length(transient_array);
    temp_index++){
  set_matrix_cell(*address_index, temp_index,
      get_array_item(temp_index, transient_array),
      pv_lookup_matrix);
}

set_matrix_cell(start,
    stop,
    (double) *address_index,
    reference_matrix);
   free_array(transient_array);
   *address_index = *address_index + 1;
    }
    stop = stop + 1;
  } // reference matrix (for tomtom only)
} // i

// Free space.
free_array(pdf_old);

// Return the pdf.
return(pdf_new);
}


/**************************************************************************
* Processes a block of sorted tomtom output lines (selects strand).
**************************************************************************/
void select_strand(STRING_LIST_T* lines,
     STRING_LIST_T* final_output){
  int i_line;
  STRING_LIST_T* completed_ids;
  completed_ids = new_string_list();

for (i_line = 0; i_line < get_num_strings(lines); i_line++){
  int offset, overlap;
    float evalue, uncorrected_pvalue;
    char query_id[MAX_MOTIF_ID_LENGTH+1], target_id[MAX_MOTIF_ID_LENGTH+1];
    char query_consensus[MAX_MOTIF_LENGTH+1];
    char target_consensus[MAX_MOTIF_LENGTH+1];

    sscanf(get_nth_string(i_line, lines),
     "%s\t%s\t%d\t%g\t%d\t%s\t%s\t%g",
     query_id, target_id, &offset, &evalue,
     &overlap, query_consensus, target_consensus,
     &uncorrected_pvalue);

    /* Get the orientation */
    char orientation = '-';
    if (target_id[0] == '+') {
      orientation = '+';
    }

    char new_target_id[MAX_MOTIF_ID_LENGTH+1];
    sprintf(new_target_id, "%s", target_id+1);

    if (!have_string(new_target_id, completed_ids)) {
      char curr_line[MAX_LINE_LENGTH+MAX_MOTIF_ID_LENGTH+1];

      sprintf(curr_line, "%s\t%s\t%d\t%g\t%g\t%d\t%s\t%s\t%c",
        query_id, 
        new_target_id, 
        offset,
	uncorrected_pvalue,
        evalue, 
        overlap, 
        query_consensus,
        target_consensus, 
        orientation
      );

      add_string(curr_line, final_output);

      add_string(new_target_id, completed_ids);
    }
  }

  /* Free memory */
  free_string_list(completed_ids);

}/* Function select_strand_new */

/**************************************************************************
 * Gets the consensus DNA sequence from a pssm.
 **************************************************************************/
void get_cons(MATRIX_T* freqs, char** cons){
  int index_rows;
  int index_cols;
  char curr_cons[1000] = "";
  STRING_LIST_T* alph = new_string_list();

  if (get_num_cols(freqs) == 4) {
    add_string("A", alph);
    add_string("C", alph);
    add_string("G", alph);
    add_string("T", alph);
  } else {
    add_string("A", alph);
    add_string("C", alph);
    add_string("D", alph);
    add_string("E", alph);
    add_string("F", alph);
    add_string("G", alph);
    add_string("H", alph);
    add_string("I", alph);
    add_string("K", alph);
    add_string("L", alph);
    add_string("M", alph);
    add_string("N", alph);
    add_string("P", alph);
    add_string("Q", alph);
    add_string("R", alph);
    add_string("S", alph);
    add_string("T", alph);
    add_string("V", alph);
    add_string("W", alph);
    add_string("Y", alph);
  }

  for (index_rows = 0; index_rows < get_num_rows(freqs);
       index_rows++) {
    int curr_index;
    curr_index = -1;
    float max;
   max = 0;

    for (index_cols = 0; index_cols < get_num_cols(freqs);
   index_cols++) {
      if (get_matrix_cell(index_rows, index_cols, freqs) > max){
  max = get_matrix_cell(index_rows, index_cols, freqs);
  curr_index = index_cols;
      }
    }
    char * temp;
    temp = strcat(curr_cons, get_nth_string(curr_index, alph));
  }
  *cons = curr_cons;
  free_string_list(alph);
} /* Function get_cons */


/**************************************************************************
 * Computes the optimal offset and score corresponding to the smallest
 * pvalue.
 **************************************************************************/
double compare_motifs (
   MATRIX_T* pairwise_column_scores,
   int*      optimal_offset,
   MATRIX_T* reference_matrix,
   MATRIX_T* pv_lookup_matrix,
   double*   optimal_pvalue,
   char*     mode,
   int *     overlapping_bases,
   char*     distance_measure,
   BOOLEAN_T internal, // Is shorter motif contained in longer motif?
   int       min_overlap,
   int*      address_index,
   MATRIX_T* all_columnwise_scores,
   int       target_database_length,
   ARRAY_T*  bg_freqs_mod
   ){

  double SCORE = 0;
  int first_pass = 1;

  /*  Slide one profile over another */
  int total_configurations = pairwise_column_scores->num_rows
    + pairwise_column_scores->num_cols - 1;
  int current_configuration;
  int overlap = -1;
  for(current_configuration = 1;
      current_configuration <= total_configurations;
      current_configuration++){

    /* If requested, check whether one motif is contained within the other. */
    BOOLEAN_T process = TRUE;
    if (internal) {
      int max, min;
      max = pairwise_column_scores->num_rows;
      min = pairwise_column_scores->num_cols;
      if (max < min) {
	int temp;
	temp = min;
	min = max;
	max = temp;
      }
      if (!((current_configuration <= max)
	    && (current_configuration >= min))) {
	process = FALSE;
      }
    }

    /* Override min-overlap if query/target is smaller than min-overlap */
    int mo;
    mo = min_overlap;
    if (pairwise_column_scores->num_rows < mo) {
      mo = pairwise_column_scores->num_rows;
    }
    if (pairwise_column_scores->num_cols < mo) {
      mo = pairwise_column_scores->num_cols;
    }

    /* Check if this configuration should be processed (min_overlap). */
    if (mo > 1) {
      int max, min;
      min = mo;
      max = pairwise_column_scores->num_rows
	+ pairwise_column_scores->num_cols - mo;
      if (!((current_configuration <= max)
	    && (current_configuration >= min))) {
	process = FALSE;
      }
    }

    /* Check whether the requested percentage of information content is 
       contained within the matching portion of each motif. */
    /* FIXME: Stub to be filled in.  WSN 4 Apr 2008
    info_content = MIN(compute_info_content(start1, stop1, motif1),
		       compute_info_content(start2, stop2, motif2));
    if (info_content < target_info_content) {
      process = FALSE;
    }
    */

    if (process) {
      double pvalue = 1;
      double TEMP_SCORE = 0;
      int current_overlap = -1;

      /* Compute the score and pvalue for the current configuration */
      TEMP_SCORE = compute_overlap_score(
					 pairwise_column_scores,
					 current_configuration,
					 reference_matrix,
					 pv_lookup_matrix,
					 &pvalue,
					 &current_overlap,
					 mode,
					 distance_measure,
					 &(*address_index),
					 all_columnwise_scores,
					 target_database_length,
					 bg_freqs_mod
					 );

      if (current_overlap > 0) {
	/* Keep the configuration with the lowest pvalue.
	   If pvalue is the same, keep the largest overlap */
	if (first_pass == 1){
	  SCORE = TEMP_SCORE;
	  *optimal_offset = current_configuration;
	  *optimal_pvalue = pvalue;
	  overlap = current_overlap;
	  first_pass = 0;
	}
	else{
	  if (pvalue == *optimal_pvalue){
	    if (overlap < current_overlap){
	      SCORE = TEMP_SCORE;
	      *optimal_offset = current_configuration;
	      *optimal_pvalue = pvalue;
	      overlap = current_overlap;
	    }
	  }
	  else{
	    if(pvalue < *optimal_pvalue){
	      SCORE = TEMP_SCORE;
	      *optimal_offset = current_configuration;
	      *optimal_pvalue = pvalue;
	      overlap = current_overlap;
	    }
	  }
	}/* For the pvalue minimization condition */
      } /* Overlap > 0 (happens sometimes if min-overlap is overridden by
	   target motif length). */
    } /* Process */
  } /* Loop over configurations. */
  *overlapping_bases = overlap;
  return SCORE;
} /* Function compare_motifs */

/**************************************************************************
 * Computes the score for a particular configuration
 **************************************************************************/
double compute_overlap_score (MATRIX_T* pairwise_column_scores,
            int current_configuration,
            MATRIX_T* reference_matrix,
            MATRIX_T* pv_lookup_matrix,
            double *pvalue,
            int *current_overlap,
            char* mode,
            char* distance_measure,
            int* address_index,
            MATRIX_T* all_columnwise_scores,
            int target_database_length,
            ARRAY_T* bg_freqs_mod){

  /* Initialize the columns from where the comparion needs to be started */
  int current_target_column = 0;
  int current_query_column = 0;
  if (current_configuration <= pairwise_column_scores->num_rows){
    current_target_column = 0;
    current_query_column = pairwise_column_scores->num_rows
      - current_configuration;
  }
  else{
    current_query_column = 0;
    current_target_column = current_configuration
      - pairwise_column_scores->num_rows;
  }

  int count_overlap = 0;
  double temporary_score = 0;

  int index_start;
  int index_stop;
  index_start = current_query_column;

 /*   Recurse over the overlapping columns */
  do {
    double SC;
    SC = get_matrix_cell(current_query_column, current_target_column,
       pairwise_column_scores);

    temporary_score = temporary_score + SC;

    current_query_column++;
    current_target_column++;
    count_overlap++;
  } while ((current_query_column < pairwise_column_scores->num_rows)
     &&(current_target_column < pairwise_column_scores->num_cols));

  index_stop = current_query_column - 1;/* -1 for the last increment */
  *current_overlap = count_overlap;

  int curr_address_index = 0;
  curr_address_index = (int) get_matrix_cell(index_start,
               index_stop,
               reference_matrix);

  if (curr_address_index == -1) {

    MATRIX_T* subset = NULL;
    get_matrix_rows(index_start, index_stop,
        all_columnwise_scores, &subset);

    ARRAY_T* pv_lookup;
    pv_lookup = get_pv_lookup_new(
          subset,
          BINS,
          target_database_length,/* alphabet size */
          bg_freqs_mod,
          index_start,
          &(*address_index),
          reference_matrix,
          pv_lookup_matrix
          );

    free_array(pv_lookup);
    free_matrix(subset);
    curr_address_index = *address_index - 1;
  }

  /* Look up the integer score in the corresponding pvlookup array */
  int integral_score = 0;
  integral_score = (int) temporary_score;

  *pvalue = get_array_item(integral_score,
         get_matrix_row(curr_address_index,
            pv_lookup_matrix));


  /* Print out pvalues for position qq plot */
  if (strcmp(mode, "qq") == 0) {
    printf("%g \n", *pvalue);
  }
  return(temporary_score);
} /* Function compute_overlap_score */

/**************************************************************************
 * Computes ALLR scores for all columns of the query matrix against target
 * motif. This can then be used to by compute_allr.
 **************************************************************************/
void allr_scores(MOTIF_T* query_motif,
     MOTIF_T* target_motif,
     ARRAY_T* background,
     MATRIX_T** score,
     char* mode){

  /*   Recurse pairwise over the columns */
  int index_query;
  int index_target;
  for (index_query = 0; index_query < query_motif->length;
       index_query++) {
    for (index_target = 0; index_target < target_motif->length;
   index_target++) {

      double SCORE_F = 0;
      double SCORE_NUMERATOR = 0;
      double SCORE_DENOMINATOR = 0;
      int index; /* Index corresponds to the alphabets. */

      for (index = 0; index < query_motif->alph_size; index++) {
  /* The numerator of the ALLR */
 double nr_part1, nr_part2;
  /*  Part 1 of the numerator  */
  nr_part1 = (((target_motif->num_sites)*
    (get_matrix_cell(index_target, index,
        target_motif->freqs))) /* nb  for target */
   * log((get_matrix_cell(index_query, index,
             query_motif->freqs))
         / get_array_item(index,
        background))); /* Likelihood for query */
   /*  Part 2 of the numerator */
  nr_part2 = (((query_motif->num_sites)*
         (get_matrix_cell(index_query, index,
              query_motif->freqs))) /* nb  for query */
        * log((get_matrix_cell(index_target, index,
             target_motif->freqs))
        / get_array_item(index,
             background))); /* Likelihood for target */
   /* Sum of the two parts. */
 SCORE_NUMERATOR += nr_part1 + nr_part2;
      /*  The denominoator of the ALLR */
      SCORE_DENOMINATOR += (
          ((query_motif->num_sites)
           *  (get_matrix_cell(index_query,
            index,
            query_motif->freqs)))/* nb for query */
          + ((target_motif->num_sites)
             * (get_matrix_cell(index_target,
             index,
             target_motif->freqs)))/* nb for target */
          ); /* Sum of nb for query and nb for target */

      } /* Nr and Dr of the score summed over the entire alphabet. */
      SCORE_F = SCORE_NUMERATOR / SCORE_DENOMINATOR;
      set_matrix_cell(index_query, index_target, SCORE_F, *score);

      /* Print out to analyze dependence on number of sites */
      if (strcmp(mode, "dep") == 0) {
        printf("%s\t%s\t%d\t%d\t%g\t%f\t%f\n", query_motif->id,
          target_motif->id, index_query, index_target, SCORE_F,
          query_motif->num_sites, target_motif->num_sites);
      }
    } /* Target motif */
  } /* Query motif */
} /* Function ALLR scores */

 /**************************************************************************
 * Computes euclidian distance between all columns of the query matrix
 * This score is not normalized on length and cannot be used without a
 * normalization criterion as our DP based p-value approach.
 **************************************************************************/
void ed_scores(MOTIF_T* query_motif,
    MOTIF_T* target_motif,
    ARRAY_T* background,
    MATRIX_T** score,
    char* mode){

  /*   Recurse pairwise over the columns */
  int index_query;
  int index_target;
  for (index_query = 0; index_query < query_motif->length;
       index_query++) {
    for (index_target = 0; index_target < target_motif->length;
   index_target++) {

      double SCORE_T = 0;

      int index; /* Index corresponds to the alphabets. */
      for (index = 0; index < query_motif->alph_size; index++) {

  double query_freq = get_matrix_cell(index_query,
              index,
              query_motif->freqs);
  double target_freq = get_matrix_cell(index_target,
               index,
               target_motif->freqs);
  /* Euclidean distance */
  double sq;
  sq = pow((query_freq - target_freq), 2);
  SCORE_T = SCORE_T + sq;
      }
      double SCORE_F;
      SCORE_F = - sqrt(SCORE_T);
      set_matrix_cell(index_query, index_target, SCORE_F, *score);
      /* Print out to analyze dependence on number of sites */
      if (strcmp(mode, "dep") == 0) {
        printf("%s\t%s\t%d\t%d\t%g\t%f\t%f\n", query_motif->id,
          target_motif->id, index_query, index_target, SCORE_F,
          query_motif->num_sites, target_motif->num_sites);
      }
    } /* Target motif */
  } /* Query motif */
} /* Function ed_scores */

 /**************************************************************************
 * Computes sandelin distance between all columns of the query matrix
 * This score is not normalized on length and cannot be used without a
 * normalization criterion as our DP based p-value approach.
 **************************************************************************/
void sandelin_scores(MOTIF_T* query_motif,
         MOTIF_T* target_motif,
         ARRAY_T* background,
         MATRIX_T** score,
         char* mode){

  /*   Recurse pairwise over the columns */
  int index_query;
  int index_target;
  for (index_query = 0; index_query < query_motif->length;
       index_query++) {
    for (index_target = 0; index_target < target_motif->length;
   index_target++) {

      double SCORE_T = 0;

      int index; /* Index corresponds to the alphabets. */
      for (index = 0; index < query_motif->alph_size; index++) {

  double query_freq = get_matrix_cell(index_query,
              index,
              query_motif->freqs);
  double target_freq = get_matrix_cell(index_target,
               index,
               target_motif->freqs);
  /* Euclidean distance */
  double sq;
  sq = pow((query_freq - target_freq), 2);
  SCORE_T = SCORE_T + sq;
      }
      double SCORE_F;
      SCORE_F = 2 - SCORE_T;
      set_matrix_cell(index_query, index_target, SCORE_F, *score);
      /* Print out to analyze dependence on number of sites */
      if (strcmp(mode, "dep") == 0) {
        printf("%s\t%s\t%d\t%d\t%g\t%f\t%f\n", query_motif->id,
          target_motif->id, index_query, index_target, SCORE_F,
          query_motif->num_sites, target_motif->num_sites);
      }
    } /* Target motif */
  } /* Query motif */
} /* Function sandelin*/

/**************************************************************************
 * Computes Kullback-Leiber for all columns of the query matrix against target
 * motif. This can then be used to by compute_overlap_scores. Note we use
 * a modified Kullback-leiber for computing single column scores. We do not
 * divide the score by the width. This division is replaced by our DP based
 * p-value computation.
 **************************************************************************/
void kullback_scores(MOTIF_T* query_motif,
         MOTIF_T* target_motif,
         ARRAY_T* background,
         MATRIX_T** score,
         char* mode){
  /*   Recurse pairwise over the columns */
  int index_query;
  int index_target;
  for (index_query = 0; index_query < query_motif->length;
       index_query++) {
    for (index_target = 0; index_target < target_motif->length;
   index_target++) {
      double SCORE_F = 0;
      int index; /* Index corresponds to the alphabets. */
      for (index = 0; index < query_motif->alph_size; index++) {
  double query_freq = get_matrix_cell(index_query,
              index,
              query_motif->freqs);
  double target_freq = get_matrix_cell(index_target,
               index,
               target_motif->freqs);
  /* Average Kullback */
  double avg_kull;
  avg_kull = ((query_freq * log10(query_freq/target_freq))
    + (target_freq * log10(target_freq/query_freq))) / 2;
  SCORE_F = SCORE_F + avg_kull;
      }
      SCORE_F = SCORE_F * -1;
      set_matrix_cell(index_query, index_target, SCORE_F, *score);
      /* Print out to analyze dependence on number of sites */
      if (strcmp(mode, "dep") == 0) {
        printf("%s\t%s\t%d\t%d\t%g\t%f\t%f\n", query_motif->id,
          target_motif->id, index_query, index_target, SCORE_F,
          query_motif->num_sites, target_motif->num_sites);
      }
    } /* Target motif */
  } /* Query motif */
} /* Function Kullback*/


/**************************************************************************
 * Computes Pearson scores for all columns of the query matrix against target
 * motif. This can then be used to by compute_overlap_scores.
 **************************************************************************/
void pearson_scores(MOTIF_T* query_motif,
        MOTIF_T* target_motif,
        ARRAY_T* background,
        MATRIX_T** score,
        char* mode){

   /*   Define the average values */
   double x_bar = 1 / query_motif->alph_size;
   double y_bar = 1 / target_motif->alph_size;

  /*   Recurse pairwise over the columns */
  int index_query;
  int index_target;
  for (index_query = 0; index_query < query_motif->length;
       index_query++) {
    for (index_target = 0; index_target < target_motif->length;
   index_target++) {

      double SCORE_F = 0;
      double SCORE_NUMERATOR = 0;
      double SCORE_SQ_DENOMINATOR_1 = 0;
      double SCORE_SQ_DENOMINATOR_2 = 0;

      int index; /* Index corresponds to the alphabets. */
      for (index = 0; index < query_motif->alph_size; index++) {
  /* The numerator of the Pearson */
  SCORE_NUMERATOR += (get_matrix_cell(index_query,
              index,
              query_motif->freqs) - x_bar)
  * (get_matrix_cell(index_target,
         index,
         target_motif->freqs) - y_bar);

  /*  The denominoator of the Pearson */
  SCORE_SQ_DENOMINATOR_1 += pow((get_matrix_cell(index_query,
              index,
              query_motif->freqs) - x_bar),
              2);
  SCORE_SQ_DENOMINATOR_2 += pow((get_matrix_cell(index_target,
              index,
              target_motif->freqs) - y_bar),
              2);
      } /* Nr and Dr components summed over the entire alphabet. */

      /*       Compute the Denominator */
      double SCORE_SQ_DENOMINATOR = SCORE_SQ_DENOMINATOR_1
  * SCORE_SQ_DENOMINATOR_2;
      double SCORE_DENOMINATOR = sqrt(SCORE_SQ_DENOMINATOR);

      SCORE_F = SCORE_NUMERATOR / SCORE_DENOMINATOR;
      set_matrix_cell(index_query, index_target, SCORE_F, *score);
      /* Print out to analyze dependence on number of sites */
      if (strcmp(mode, "dep") == 0) {
        printf("%s\t%s\t%d\t%d\t%g\t%f\t%f\n", query_motif->id,
          target_motif->id, index_query, index_target, SCORE_F,
          query_motif->num_sites, target_motif->num_sites);
      }
    } /* Target motif */
  } /* Query motif */
} /* Function pearson */

// FIXME: currently prints HTML
void print_tomtom_xml_header(
  FILE *output_path,
  char *query_filename,
  char *target_filename,
  char *distance_measure,
  float ethresh
) 
{
  fprintf(output_path,
    "<html>\n"
    "<head>\n"
    "  <title> Tomtom output</title>\n"
    "</head> <body>\n" 
    "<table border=1>\n"
    "  <tr> <th align=center colspan=2><h2>TOMTOM OUTPUT</h2></th></tr>\n"
    "  <tr> <td align=left colspan=2><b>Query File:</b> %s</td></tr>\n"
    "  <tr> <td align=left colspan=2><b>Target File:</b> %s</td></tr>\n"
    "  <tr> <td align=left colspan=2><b>Distance Measure:</b> %s</td></tr>\n"
    "  <tr> <td align=left colspan=2><b>All Motif Matches with <i>E</i>-value at most:</b> %.2g</td></tr>\n",
    query_filename,
    target_filename,
    distance_measure,
    ethresh
  );
} // print_tomtom_xml_header

// FIXME: currently prints HTML
void print_tomtom_xml_line(
  FILE *output_path,		// path to output XML file
  char *line,			// stored results line
  char *output_dirname,		// name of output directory
  MOTIF_T *motifs_query,	// query motifs
  MOTIF_T *motifs_target,	// target motifs
  char *query_url_format,	// format string for query URL
  char *target_url_format 	// format string for target URL
  ) 
{
  int query, target, offset, overlap;
  char query_id[MAX_MOTIF_ID_LENGTH+1], target_id[MAX_MOTIF_ID_LENGTH+1];
  double evalue, uncorrected_pvalue;
  char orientation;

  // Parse the stored results line; 
  // must match sprintf in select_strand()
  sscanf(line, "%d,%d,%d|%s\t%s\t%*d\t%lg\t%lg\t%d\t%*s\t%*s\t%c",
    &query,
    &target,
    &offset,
    query_id, 
    target_id, 
    // offset_by_length,  // suppressed
    &uncorrected_pvalue,
    &evalue,
    &overlap, 
    // query_cons,	// suppressed
    // get_nth_string(motif_index_target, motif_cons_target), // supr.
    &orientation
  );

  // create alignment logo
  char *logo_filename = create_logo( 
    &(motifs_target[target]), 
    &(motifs_query[query]),
    offset, 
    output_dirname
  );

  // create href for motifs by using the given format
  char *query_href = (char*) malloc(
    sizeof(char) * (strlen(query_url_format) + 2*strlen(query_id) + 1)
  );
  sprintf(query_href, query_url_format, query_id, query_id);
  char *target_href = (char*) malloc(
    sizeof(char) * (strlen(target_url_format) + 2*strlen(target_id) + 1)
  );
  sprintf(target_href, target_url_format, target_id, target_id);

  //FIXME: change to XML
  fprintf(output_path,
    "  <tr>\n"
    "    <td>\n"
    "      <table>\n"
    "        <tr><th align=left nowrap>Target Motif:</th><td align=right>%s</td></tr>\n"	// target id
    "        <tr><th align=left nowrap>Target Description:</th><td align=right>%s</td></tr>\n"	// target description
    "        <tr><th align=left nowrap>Query Motif:</th><td align=right>%s</td></tr>\n"		// query id
    "        <tr><th align=left nowrap>Query Description:</th><td align=right>%s</td></tr>\n"	// query description
    "        <tr><th align=left nowrap><i>p</i>-value:</th><td align=right><b>%.2g<b></td></tr>\n"	// p-value
    "        <tr><th align=left nowrap><i>E</i>-value:</th><td align=right><b>%.2g<b></td></tr>\n"	// E-value
    "        <tr><th align=left nowrap>Overlap:</th><td align=right>%d</td></tr>\n"		// overlap
    "        <tr><th align=left nowrap>Query Offset:</th><td align=right>%d</td></tr>\n"	// offset
    "        <tr><th align=left nowrap>Orientation:</th><td align=right>%c</td></tr>\n"		// orientation
    "        <tr><th align=left nowrap>Figures:</th>"						// links
    "<td align=right><a href='%s.eps'>[EPS]</a><a href='%s.png'>[PNG]</a></td></tr>\n"
    "      </table>\n"
    "    </td>\n"
    "    <td><img src='%s.png'></td>\n"							// logo
    "  </tr>\n",
    target_href, 
    motifs_target[target].id2,			// target description
    query_href, 
    motifs_query[query].id2,			// query description
    uncorrected_pvalue,
    evalue,
    overlap,
    offset,
    orientation,
    logo_filename, logo_filename,		// links
    logo_filename 				// image
  );

  myfree(logo_filename);
  myfree(query_href);
  myfree(target_href);
} // print_tomtom_xml_line

// FIXME: currently prints HTML
void print_tomtom_xml_tailer(
  FILE *output_path,
  int argc,
  char **argv,
  int total_matches
) 
{
  //FIXME: change to XML
  int i;

  fprintf(output_path, 
    "  <tr>\n"
    "    <td> <b>Total matches under <i>E-</i>value threshold:</b> %d\n"
    "    </td>\n"
    "  </tr>\n", 
    total_matches);
  fprintf(output_path, "<tr><th>Command line:</th><td>\n");
  for (i=0; i<argc; i++) {
    fprintf(output_path, "%s\n", argv[i]);
  }
  fprintf(output_path, "</td></tr>\n");
  fprintf(output_path,
    "</table>\n\n"
    "</body>\n"
    "</html>\n"
  );
} // print_tomtom_xml_tailer

/*
 * Local Variables:
 * mode: c
 * c-basic-offset: 2
 * End:
 */
