/****************************************************************************
 * FILE: tomtom.h
 * AUTHOR: Shobhit Gupta
 * CREATE DATE: 02/09/2006
 * PROJECT: TOMTOM
 * DESCRIPTION: Compare two motifs using the ALLR statistic
 * COPYRIGHT: 2006, UW
 ****************************************************************************/
#include "matrix.h"
#include "array.h"
#include "motif.h"
#include "string-list.h"
#include "utils.h"

/**************************************************************************
*
* get_pv_lookup_new
*
* Create a lookup table for motif scores.
*
* Returns an array of pdf values:
*   pv[x] = Pr(score >= x)
* for 0<=x<=range*w.
*
* Assumes:
*   1) motif scores are non-negative, integral
*   2) background model is 0-order Markov
*
*
* The last four parameters are specific to tomtom (described below).
* For all other usege, these frour parameters can be passed as
* NULL, -1,  NULL, NULL.
*
* Four additional parameters for tomtom:
* Four additional parameters are added for tomtom to store the
* intermediate arrays. For this purpose, a pvalue matrix is passed
* on together with a reference matrix. One integer pointer is passed
* that stores the last row number of the p-value matrix. Another integer
* is passed to report which query column to start with
*
**************************************************************************/
ARRAY_T* get_pv_lookup_new(
  MATRIX_T* pssm,     // The PSSM.
  int range,          // Maximum score in PSSM.
  int alen,           // Alphabet size.
  ARRAY_T* background,  // The background model.
  int, //Start position of query motif column (tomtom only)
  int *, // index: last row of pv_lookup_matrix (tomtom only)
  MATRIX_T*, // reference_matrix (tomtom only)
  MATRIX_T* // pv_lookup_matrix (tomtom only)
  );

/**************************************************************************
*
* get_pdf_lookup_new
*
* Create a lookup table for the pdf.
*
* Returns an array of pdf values:
*   pdf[x] = Pr(score == x)
* for 0<=x<=range*w.
*
* Assumes:
*   1) motif scores are non-negative, integral
*   2) background model is 0-order Markov
*
* The last four parameters are specific to tomtom (described below).
* For all other usege, these frour parameters can be passed as
* NULL, -1,  NULL, NULL.
*
* Four additional parameters for tomtom:
* Four additional parameters are added for tomtom to store the
* intermediate arrays. For this purpose, a pvalue matrix is passed
* on together with a reference matrix. One integer pointer is passed
* that stores the last row number of the p-value matrix. Another integer
* is passed to report which query column to start with
*
**************************************************************************/
static ARRAY_T* get_pdf_lookup_new(
  MATRIX_T* pssm,       // The PSSM.
  int w,          // Width of PSSM.
  int alen,         // Length of alphabet.
  int range,          // Range of scores.
  ARRAY_T* background,        // Background model.
  int, //Start position of query motif column (tomtom only)
  int *, // index: last row of pv_lookup_matrix (tomtom only)
  MATRIX_T*, // reference_matrix (tomtom only)
  MATRIX_T* // pv_lookup_matrix (tomtom only)
  );

/**************************************************************************
 * Processes a block of sorted tomtom output lines (selects strand).
 **************************************************************************/
void select_strand(
           STRING_LIST_T*, // Source
           STRING_LIST_T* // Destination
           );

/**************************************************************************
 * Computes consensus sequence based on the given freq matrix. Works only
 * for DNA alphabets
 **************************************************************************/

void get_cons(
        MATRIX_T*,
        char** //cons sequence
        );

/**************************************************************************
 * Computes the optimal offset and ALLR score corresponding to the smallest
 + pvalues.
 **************************************************************************/

double compare_motifs(
          MATRIX_T*,//computed_scores
          int*,//optimal_offset
          MATRIX_T*,//reference matrix for pv lookup
          MATRIX_T*,//pv lookup matrix
          double*,//optimal pvalue
          char*, //Program mode for print options :qq/dep
          int*, //Number of Overlapping bases
          char*, //Distance measure
          BOOLEAN_T, //internal alignments
          int, // min_overlap
          int *,//Address index
          MATRIX_T*,
          int,
          ARRAY_T*
          );
/**************************************************************************
 * Computes the score for a particular configuration
 **************************************************************************/

double compute_overlap_score(
           MATRIX_T*,//computed_scores
           int,//current_configuration
           MATRIX_T*,//reference matrix for pv lookup
           MATRIX_T*,//pv lookup matrix
           double*,//pvalue
           int*,//overlap between query and target
           char*, //Program mode for print options :qq/dep
           char*, //Distance measure
           int *, //Address index
           MATRIX_T*,
           int,
           ARRAY_T*
           );

/**************************************************************************
 * Computes sandelin scores for all columns of the query matrix against
 * target motif. This can then be used to by compute_overlap_score.
 **************************************************************************/

void sandelin_scores(
         MOTIF_T*,//query_motif
         MOTIF_T*,//target_motif
         ARRAY_T*,//background_frequencies
         MATRIX_T**,//output score matrix
         char* //Program mode for print options :qq/dep
         );

/**************************************************************************
 * Computes Euclidean scores for all columns of the query matrix against target
 * motif. This can then be used to by compute_overlap_score.
 **************************************************************************/

void ed_scores(
    MOTIF_T*,//query_motif
    MOTIF_T*,//target_motif
    ARRAY_T*,//background_frequencies
    MATRIX_T**,//output score matrix
    char* //Program mode for print options :qq/dep
    );

/**************************************************************************
 * Computes Kullback scores for all columns of the query matrix against target
 * motif. This can then be used to by compute_overlap_score.
 **************************************************************************/

void kullback_scores(
         MOTIF_T*,//query_motif
         MOTIF_T*,//target_motif
         ARRAY_T*,//background_frequencies
         MATRIX_T**,//output score matrix
         char* //Program mode for print options :qq/dep
         );

/**************************************************************************
 * Computes ALLR scores for all columns of the query matrix against target
 * motif. This can then be used to by compute_overlap_score.
 **************************************************************************/

void allr_scores(
        MOTIF_T*,//query_motif
        MOTIF_T*,//target_motif
        ARRAY_T*,//background_frequencies
        MATRIX_T**,//output score matrix
        char* //Program mode for print options :qq/dep
        );

/**************************************************************************
 * Computes Pearson scores for all columns of the query matrix against target
 * motif. This can then be used to by compute_overlap_score.
 **************************************************************************/

void pearson_scores(
        MOTIF_T*,//query_motif
        MOTIF_T*,//target_motif
        ARRAY_T*,//background_frequencies
        MATRIX_T**,//output score matrix
        char* //Program mode for print options :qq/dep
        );

/*
 * Local Variables:
 * mode: c
 * c-basic-offset: 2
 * End:
 */
